from keras_cv_attention_models.nfnets.nfnets import (
    NormFreeNet,
    NFNetF0,
    NFNetF1,
    NFNetF2,
    NFNetF3,
    NFNetF4,
    NFNetF5,
    NFNetF6,
    NFNetF7,
    NormFreeNet_Light,
    NFNetL0,
    ECA_NFNetL0,
    ECA_NFNetL1,
    ECA_NFNetL2,
    ECA_NFNetL3,
    ScaledStandardizedConv2D,
    ZeroInitGain,
)


__head_doc__ = """
Keras implementation of [NFNets](https://github.com/deepmind/deepmind-research/tree/master/nfnets).
Paper [PDF 2102.06171 High-Performance Large-Scale Image Recognition Without Normalization](https://arxiv.org/pdf/2102.06171.pdf).
"""

__tail_doc__ = """  stem_width: output dimension for stem block. Default 128.
  out_channels: default as `[256, 512, 1536, 1536]`. Output channel for each stack.
  channel_ratio: filter scale ratio for block hidden layers. `0.5` for `F` original models, `0.25` for `L` light models.
  num_features_factor: none `0` value to added a `ScaledStandardizedConv2D` layer before output,
      added layer `filters = num_features_factor * out_channels[-1] * width_factor`.
      Default `1.5` for `L0` models, `2` for others.
  strides: list value indicates strides used in the each stack. Default `[1, 2, 2, 2]`.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
      Set `(None, None, 3)` for dynamic input resolution.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  se_ratio: value in `(0, 1)`, where `0` means not using `se_module`, indicates `se_ratio` for all blocks.
      Default `0.5` for `F` original models, `0.25` for `L` light models.
  group_size: groups size for all blocks 3x3 Conv2D layers, `groups = filters // group_size`.
      Default `128` for `F` original models, `64` for `L` light models.
  use_zero_init_gain: boolean value if add a `ZeroInitGain` layer for deep branch in all blocks.
      Default `True` for `F` original models, `False` for `L` light models.
  torch_padding: boolean value if using PyTorch like `padding`. Default `False` for `F` original models, `True` for `L` light models.
  gamma_in_act: boolean value if using non linear gamma in activation or conv2d layer.
      Default `True` for `F` original models, `False` for `L` light models.
  alpha: scale factor for deep branch in all blocks. Currently all using `0.2`.
  width_factor: width expansion for entire model. Currently all using `1.0`.
  activation: activation used in whole model.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  dropout: dropout rate if top layers is included.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

NormFreeNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  attn_type: attention types for all blocks, could be `se` or `eca`. Default is `se`.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model       | Params | FLOPs   | Input | Top1 Acc |
  | ----------- | ------ | ------- | ----- | -------- |
  | NFNetL0     | 35.07M | 7.13G   | 288   | 82.75    |
  | NFNetF0     | 71.5M  | 12.58G  | 256   | 83.6     |
  | NFNetF1     | 132.6M | 35.95G  | 320   | 84.7     |
  | NFNetF2     | 193.8M | 63.24G  | 352   | 85.1     |
  | NFNetF3     | 254.9M | 115.75G | 416   | 85.7     |
  | NFNetF4     | 316.1M | 216.78G | 512   | 85.9     |
  | NFNetF5     | 377.2M | 291.73G | 544   | 86.0     |
  | NFNetF6 SAM | 438.4M | 379.75G | 576   | 86.5     |
  | NFNetF7     | 499.5M | 481.80G | 608   |          |
  | ECA_NFNetL0 | 24.14M | 7.12G   | 288   | 82.58    |
  | ECA_NFNetL1 | 41.41M | 14.93G  | 320   | 84.01    |
  | ECA_NFNetL2 | 56.72M | 30.12G  | 384   | 84.70    |
  | ECA_NFNetL3 | 72.04M | 52.73G  | 448   |          |
"""

NFNetF0.__doc__ = __head_doc__ + """
Args:
  attn_type: attention types for all blocks, could be `se` or `eca`. Default is `se`.
""" + __tail_doc__

NFNetF1.__doc__ = NFNetF0.__doc__
NFNetF2.__doc__ = NFNetF0.__doc__
NFNetF3.__doc__ = NFNetF0.__doc__
NFNetF4.__doc__ = NFNetF0.__doc__
NFNetF5.__doc__ = NFNetF0.__doc__
NFNetF6.__doc__ = NFNetF0.__doc__
NFNetF7.__doc__ = NFNetF0.__doc__

__light_head_doc__ = __head_doc__ + """Light versions of `NFNetF` from `timm`.
Weights reloaded from [Github rwightman/pytorch-image-models](https://github.com/rwightman/pytorch-image-models).
"""

NormFreeNet_Light.__doc__ = __light_head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  attn_type: attention types for all blocks, could be `se` or `eca`. Default is `se`.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model       | Params | FLOPs   | Input | Top1 Acc |
  | ----------- | ------ | ------- | ----- | -------- |
  | NFNetL0     | 35.07M | 7.13G   | 288   | 82.75    |
  | ECA_NFNetL0 | 24.14M | 7.12G   | 288   | 82.58    |
  | ECA_NFNetL1 | 41.41M | 14.93G  | 320   | 84.01    |
  | ECA_NFNetL2 | 56.72M | 30.12G  | 384   | 84.70    |
  | ECA_NFNetL3 | 72.04M | 52.73G  | 448   |          |
"""

NFNetL0.__doc__ = __light_head_doc__ + """
Args:
  attn_type: attention types for all blocks, could be `se` or `eca`. Default is `se`.
""" + __tail_doc__

ECA_NFNetL0.__doc__ = __light_head_doc__ + """Model Using `attn_type="eca"` instead of `attn_type="se"`.

Args:
""" + __tail_doc__

ECA_NFNetL1.__doc__ = ECA_NFNetL0.__doc__
ECA_NFNetL2.__doc__ = ECA_NFNetL0.__doc__
ECA_NFNetL3.__doc__ = ECA_NFNetL0.__doc__

ScaledStandardizedConv2D.__doc__ = __head_doc__ + """
Scaled Standardized Conv2D layer, sub-class of typical Conv2D.

Additional args from Conv2D:
  gamma: is a scaling factor multipled to scale.
  eps: is small constant used to avoid dividing 0.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.ScaledStandardizedConv2D(filters=64, kernel_size=3)
>>> print(f"{aa(tf.ones([1, 28, 28, 32])).shape = }")
# aa(tf.ones([1, 28, 28, 32])).shape = TensorShape([1, 26, 26, 64])
>>> print({ii.name:ii.shape for ii in aa.weights})
# {'scaled_standardized_conv2d/kernel:0': TensorShape([3, 3, 32, 64]),
#  'scaled_standardized_conv2d/bias:0': TensorShape([64]),
#  'scaled_standardized_conv2d/gain:0': TensorShape([64])}
"""

ZeroInitGain.__doc__ = __head_doc__ + """
Zero Init Gain.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.ZeroInitGain()
>>> print(f"{aa(tf.ones([1, 28, 28, 32])).shape = }")
# aa(tf.ones([1, 28, 28, 32])).shape = TensorShape([1, 26, 26, 64])
>>> print({ii.name:ii.shape for ii in aa.weights})
# {'zero_init_gain/gain:0': TensorShape([])}
>>> print(aa.gain.numpy())
# 0.0
"""



import os
import numpy as np
from PIL import Image
from tqdm.auto import tqdm
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import callbacks, functional


class RunPrediction:
    def __init__(self, model, num_training_steps=1000, num_steps=0, beta_max=0.02):
        self.model, self.num_training_steps = model, num_training_steps
        self.num_steps = num_steps if num_steps > 0 else num_training_steps

        self.timesteps = np.arange(0, num_training_steps, num_training_steps // self.num_steps).astype("int64")
        beta = np.linspace(0.0001, beta_max, num_training_steps).astype("float32")[self.timesteps]
        alpha = 1.0 - beta
        alpha_bar = np.cumprod(alpha, axis=0)
        self.xt_prev_alpha = 1 / (alpha**0.5)
        self.xt_noise_alpha = self.xt_prev_alpha * (1 - alpha) / (1 - alpha_bar) ** 0.5
        self.eps_alpha = beta**0.5

        self.bar_format = "{desc}: {n_fmt}/{total_fmt} - [{elapsed}<{remaining},{rate_fmt}]"
        self.is_channels_last = backend.image_data_format() == "channels_last"
        self.device = None  # Set to actual torch model using device later
        self.to_device = (lambda xx: xx.to(self.device)) if backend.is_torch_backend else (lambda xx: xx)
        self.to_host = (lambda xx: xx.cpu()) if backend.is_torch_backend else (lambda xx: xx)

    def __call__(self, image_size=-1, batch_size=1, labels=None, init_x0=None, init_step=0, labels_guide_weight=1.8, return_inner=False):
        assert (len(self.model.inputs) == 2 and labels is None) or (len(self.model.inputs) == 3 and labels is not None), "labels input not matching with model"

        if backend.is_torch_backend and self.device is None:
            self.device = next(self.model.parameters()).device

        """ Init labels inputs first, for getting batch_size by labels """
        use_labels = False if labels is None else True
        if use_labels:
            if init_x0 is not None:
                assert init_x0.shape[0] == len(labels), "init_x0.shape[0] = {} should be equal with len(labels) = {}".format(init_x0.shape[0], len(labels))
            batch_size = len(labels)
            labels_inputs = self.to_device(functional.convert_to_tensor(labels, dtype="int64")) + 1  # add 1 skipping 0
            labels_inputs_zeros = self.to_device(functional.convert_to_tensor(np.zeros([batch_size]), dtype="int64"))

        """ Create init_x0 """
        if init_x0 is None:
            model_input_shape = self.model.input_shape[0][1:]
            image_size = image_size if None in model_input_shape else (model_input_shape[:-1] if self.is_channels_last else model_input_shape[1:])
            image_size = image_size if isinstance(image_size, (list, tuple)) else (image_size, image_size)
            noise_shape = (batch_size, *image_size, 3) if self.is_channels_last else (batch_size, 3, *image_size)
            init_x0 = np.random.normal(size=noise_shape).astype("float32")
        else:
            noise_shape, batch_size = init_x0.shape, init_x0.shape[0]
        xt = self.to_device(functional.convert_to_tensor(init_x0, dtype="float32"))

        """ Unet loop steps """
        gathered_inner = []
        for cur_step in tqdm(range(self.num_steps - init_step)[::-1], "Eval", bar_format=self.bar_format):  # 1000 -> 0
            timestep_inputs = functional.convert_to_tensor(np.stack([self.timesteps[cur_step]] * batch_size), dtype="int64")
            timestep_inputs = self.to_device(timestep_inputs)
            if use_labels:
                xt_noise_cond = self.model([xt, labels_inputs, timestep_inputs])
                xt_noise_zeros = self.model([xt, labels_inputs_zeros, timestep_inputs])
                xt_noise = xt_noise_cond + labels_guide_weight * (xt_noise_cond - xt_noise_zeros)
            else:
                xt_noise = self.model([xt, timestep_inputs])

            eps = functional.convert_to_tensor(np.random.normal(size=noise_shape).astype("float32"), dtype=xt_noise.dtype)
            xt = self.xt_prev_alpha[cur_step] * xt - self.xt_noise_alpha[cur_step] * xt_noise
            xt += self.eps_alpha[cur_step] * self.to_device(eps)

            if return_inner:
                gathered_inner.append(xt)

        """ Output """
        if return_inner:
            gathered_inner = np.stack([self.to_host(inner).numpy().astype("float32") for inner in gathered_inner], axis=1)
            eval_xt = gathered_inner if self.is_channels_last else gathered_inner.transpose([0, 1, 3, 4, 2])  # [batch, innner_steps, height, width, channel]
        else:
            xt = self.to_host(xt).numpy().astype("float32")
            eval_xt = xt if self.is_channels_last else xt.transpose([0, 2, 3, 1])
        eval_xt = (np.clip(eval_xt / 2 + 0.5, 0, 1) * 255).astype("uint8")
        return eval_xt


class DenoisingEval(callbacks.Callback):
    def __init__(
        self, save_path, image_size=512, num_classes=0, num_training_steps=1000, num_steps=0, labels_guide_weight=1.8, beta_max=0.02, interval=1, rows=5, cols=4
    ):
        super().__init__()
        self.save_path, self.image_size, self.labels_guide_weight, self.interval = save_path, image_size, labels_guide_weight, max(interval, 1)
        self.num_classes, self.rows, self.cols, self.batch_size = num_classes, rows, cols, rows * cols
        self.run_prediction = RunPrediction(model=None, num_training_steps=num_training_steps, num_steps=num_steps, beta_max=beta_max)

        if backend.image_data_format() == "channels_last":
            self.eval_x0 = np.random.normal(size=(self.batch_size, image_size, image_size, 3)).astype("float32")
        else:
            self.eval_x0 = np.random.normal(size=(self.batch_size, 3, image_size, image_size)).astype("float32")

        if num_classes > 0:
            labels_inputs = np.arange(0, num_classes, num_classes / self.batch_size)[: self.batch_size].astype("int64")
            self.labels_inputs = np.pad(labels_inputs, [0, self.batch_size - labels_inputs.shape[0]], constant_values=num_classes - 1)  # Just in case
            print(">>>> Eval labels:", self.labels_inputs)

    def on_epoch_end(self, cur_epoch=0, logs=None):
        if (cur_epoch + 1) % self.interval > 0:
            return
        if self.run_prediction.model is None:
            self.run_prediction.model = self.model
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path, exist_ok=True)

        lables = self.labels_inputs if self.num_classes > 0 else None
        eval_xt = self.run_prediction(labels=lables, init_x0=self.eval_x0, init_step=0, labels_guide_weight=self.labels_guide_weight)
        eval_xt = np.vstack([np.hstack(eval_xt[row * self.cols : (row + 1) * self.cols]) for row in range(self.rows)])
        save_path = os.path.join(self.save_path, "epoch_{}.jpg".format(cur_epoch + 1) if isinstance(cur_epoch, int) else (cur_epoch + ".jpg"))
        Image.fromarray(eval_xt).save(save_path)
        print(">>>> Eval result image saved to {}".format(save_path))



"""
Raw ultralytics data loader
"""
import torch
from torch.utils.data import DataLoader
from ultralytics.cfg import get_cfg
from ultralytics.utils import DEFAULT_CFG
from ultralytics.data.utils import check_det_dataset
from ultralytics.data.dataset import YOLODataset


def to_data_loader(data, cfg, imgsz=640, mode="train", batch_size=16, rect_val=False):
    if mode == "train":
        augment, pad, shuffle, rect, batch_size = True, 0, True, False, batch_size
    else:
        augment, pad, shuffle, rect, batch_size = False, 0.5, False, rect_val, batch_size * 2
    dataset = YOLODataset(
        img_path=data["train"] if mode == "train" else data["val"],
        imgsz=imgsz,
        batch_size=batch_size,
        augment=augment,  # augmentation
        hyp=cfg,  # TODO: probably add a get_hyps_from_cfg function
        rect=rect,  # rectangular batches
        cache=None,
        single_cls=False,
        stride=32,
        pad=pad,
        data=data,
        # data=data,
        # classes=cfg.classes,
    )

    generator = torch.Generator()
    generator.manual_seed(6148914691236517205)
    collate_fn = getattr(dataset, "collate_fn", None)
    data_loader = DataLoader(
        dataset=dataset, batch_size=batch_size, shuffle=shuffle, num_workers=16, sampler=None, pin_memory=True, collate_fn=collate_fn, generator=generator
    )
    return data_loader


def get_data_loader(dataset_path="coco.yaml", cfg={}, imgsz=640, batch_size=16, rect_val=False):
    cfg = get_cfg(DEFAULT_CFG)
    cfg.data = dataset_path
    cfg.imgsz = imgsz
    data = check_det_dataset(dataset_path)
    train_loader = to_data_loader(data, cfg, imgsz=imgsz, mode="train", batch_size=batch_size)
    val_loader = to_data_loader(data, cfg, imgsz=imgsz, mode="val", batch_size=batch_size, rect_val=rect_val)
    return train_loader, val_loader


if __name__ == "__main__":
    os.environ["KECAM_BACKEND"] = "torch"
    sys.path.append("../ultralytics/")
    from keras_cv_attention_models.yolov8.data import get_data_loader

    train_loader, val_loader = get_data_loader("coco128.yaml")
    for aa in train_loader:
        break
    plt.imshow(aa["img"][0].permute(1, 2, 0).numpy())



import math
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format, initializers, register_keras_serializable
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    fold_by_conv2d_transpose,
    CompatibleExtractPatches,
    add_pre_post_process,
)


BATCH_NORM_EPSILON = 1e-5
PRETRAINED_DICT = {
    "volo_d1": {"imagenet": {224: "b642d39b05da9f460035d5d5fa617774", 384: "c7632a783d43278608d84f9463743b2e"}},
    "volo_d2": {"imagenet": {224: "19c6c49d3a1020e9fafbcce775200e30", 384: "fc0435d59925e547d9003010a51e4a16"}},
    "volo_d3": {"imagenet": {224: "42ae5c1be8ceb644d4f7c3d894a0034f", 448: "62304a047f182265617c49f74991e6a0"}},
    "volo_d4": {"imagenet": {224: "b45c6518b5e7624b0f6a61f18a5a7bae", 448: "c3e48df2a555032608d48841d2f4a551"}},
    "volo_d5": {"imagenet": {224: "19c98591fb2a97c2a51d9723c2ff6e1d", 448: "6f9858b667cfef77339901c3121c85a1", 512: "f2aa0cb8e265cabee840a6b83858d086"}},
}


def outlook_attention(inputs, embed_dim, num_heads=8, kernel_size=3, padding=1, strides=2, attn_dropout=0, output_dropout=0, name=""):
    _, height, width, channel = inputs.shape
    # qk_scale = float(1.0 / functional.sqrt(functional.cast(embed_dim // num_heads, "float32")))
    qk_scale = 1.0 / (float(embed_dim // num_heads) ** 0.5)
    hh, ww = int(math.ceil(height / strides)), int(math.ceil(width / strides))

    vv = layers.Dense(embed_dim, use_bias=False, name=name + "v")(inputs)

    """ attention """
    # [1, 14, 14, 192]
    pool_padding = "valid" if height % strides == 0 and width % strides == 0 else "same"
    attn = inputs if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(inputs)  # channels_last -> channels_first
    attn = layers.AvgPool2D(pool_size=strides, strides=strides, padding=pool_padding)(attn)
    attn = attn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(attn)  # channels_first -> channels_last
    # [1, 14, 14, 486]
    attn = layers.Dense(kernel_size**4 * num_heads, name=name + "attn")(attn) * qk_scale
    # [1, 14, 14, 6, 9, 9]
    attn = functional.reshape(attn, (-1, hh, ww, num_heads, kernel_size * kernel_size, kernel_size * kernel_size))
    # attention_weights = functional.softmax(attn, axis=-1)
    attention_weights = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attn)
    if attn_dropout > 0:
        attention_weights = layers.Dropout(attn_dropout)(attention_weights)

    """ unfold """
    # [1, 14, 14, 1728] if compressed else [1, 14, 14, 3, 3, 192]
    if backend.is_torch_backend:
        patches = functional.extract_patches(vv, sizes=kernel_size, strides=strides, padding="same")
    else:
        # patches = functional.extract_patches(pad_vv, patch_kernel, patch_strides, [1, 1, 1, 1], padding="valid")
        patches = CompatibleExtractPatches(kernel_size, strides, padding="same", compressed=False, name=name)(vv)

    """ matmul """
    # mm = einops.rearrange(patches, 'D H W (k h p) -> D H W h k p', h=num_head, k=kernel_size * kernel_size)
    # mm = functional.matmul(attn, mm)
    # mm = einops.rearrange(mm, 'D H W h (kh kw) p -> D H W kh kw (h p)', h=num_head, kh=kernel_size, kw=kernel_size)
    # [1, 14, 14, 9, 6, 32], the last 2 dimenions are channel 6 * 32 == 192
    mm = functional.reshape(patches, [-1, hh, ww, kernel_size * kernel_size, num_heads, embed_dim // num_heads])
    # [1, 14, 14, 6, 9, 32], meet the dimenion of attn for matmul
    mm = functional.transpose(mm, [0, 1, 2, 4, 3, 5])
    # [1, 14, 14, 6, 9, 32], The last two dimensions [9, 9] @ [9, 32] --> [9, 32]
    # mm = layers.Lambda(lambda xx: tf.matmul(xx[0], xx[1]))([attention_weights, mm])
    mm = attention_weights @ mm
    # [1, 14, 14, 9, 6, 32], transpose back
    mm = functional.transpose(mm, [0, 1, 2, 4, 3, 5])
    # [1, 14, 14, 3, 3, 192], split kernel_dimension: 9 --> [3, 3], merge channel_dimmension: [6, 32] --> 192
    mm = functional.reshape(mm, [-1, hh, ww, kernel_size, kernel_size, embed_dim])

    """ fold """
    # [1, 28, 28, 192]
    output = fold_by_conv2d_transpose(mm, inputs.shape[1:], kernel_size, strides, padding="same", compressed=False, name=name)

    # output = UnfoldMatmulFold((height, width, embed_dim), kernel_size, padding, strides)([vv, attention_weights])
    output = layers.Dense(embed_dim, use_bias=True, name=name + "out")(output)

    if output_dropout > 0:
        output = layers.Dropout(output_dropout)(output)

    return output


def outlook_attention_simple(inputs, embed_dim, num_heads=6, kernel_size=3, attn_dropout=0, name=""):
    """Simple version not using unfold and fold"""
    key_dim = embed_dim // num_heads
    # qk_scale = float(1.0 / tf.math.sqrt(tf.cast(key_dim, "float32")))
    qk_scale = 1.0 / (float(embed_dim // num_heads) ** 0.5)

    height, width = inputs.shape[1], inputs.shape[2]
    hh, ww = int(math.ceil(height / kernel_size)), int(math.ceil(width / kernel_size))  # 14, 14
    padded = hh * kernel_size - height
    if padded != 0:
        inputs = layers.ZeroPadding2D(((0, padded), (0, padded)))(inputs)

    vv = layers.Dense(embed_dim, use_bias=False, name=name + "v")(inputs)
    # vv = einops.rearrange(vv, "D (h hk) (w wk) (H p) -> D h w H (hk wk) p", hk=kernel_size, wk=kernel_size, H=num_heads, p=key_dim)
    vv = functional.reshape(vv, (-1, hh, kernel_size, ww, kernel_size, num_heads, key_dim))  # [1, 14, 2, 14, 2, 6, 32]
    vv = functional.transpose(vv, [0, 1, 3, 5, 2, 4, 6])
    vv = functional.reshape(vv, [-1, hh, ww, num_heads, kernel_size * kernel_size, key_dim])  # [1, 14, 14, 6, 4, 32]

    # attn = layers.AvgPool2D(pool_size=3, strides=2, padding='same')(inputs)
    attn = layers.AvgPool2D(pool_size=kernel_size, strides=kernel_size)(inputs)
    attn = layers.Dense(kernel_size**4 * num_heads, use_bias=True, name=name + "attn")(attn) * qk_scale
    attn = functional.reshape(attn, [-1, hh, ww, num_heads, kernel_size * kernel_size, kernel_size * kernel_size])  # [1, 14, 14, 6, 4, 4]
    # attn = functional.softmax(attn, axis=-1)
    attn = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attn)
    if attn_dropout > 0:
        attn = layers.Dropout(attn_dropout)(attn)

    # out = tf.matmul(attn, vv)  # [1, 14, 14, 6, 4, 32]
    out = attn @ vv
    # out = einops.rearrange(out, "D h w H (hk wk) p -> D (h hk) (w wk) (H p)", hk=kernel_size, wk=kernel_size)  # [1, 28, 28, 192]
    out = functional.reshape(out, [-1, hh, ww, num_heads, kernel_size, kernel_size, key_dim])  # [1, 14, 14, 6, 2, 2, 32]
    out = functional.transpose(out, [0, 1, 4, 2, 5, 3, 6])  # [1, 14, 2, 14, 2, 6, 32]
    out = functional.reshape(out, [-1, inputs.shape[1], inputs.shape[2], embed_dim])  # [1, 28, 28, 192]
    if padded != 0:
        out = out[:, :-padded, :-padded, :]
    out = layers.Dense(embed_dim, use_bias=True, name=name + "out")(out)

    return out


@register_keras_serializable(package="volo")
class BiasLayer(layers.Layer):
    def __init__(self, axis=-1, initializer="zeros", **kwargs):
        super(BiasLayer, self).__init__(**kwargs)
        self.axis, self.initializer = axis, initializer

    def build(self, input_shape):
        if self.axis == -1 or self.axis == len(input_shape) - 1:
            bb_shape = (input_shape[-1],)
        else:
            bb_shape = [1] * len(input_shape)
            axis = self.axis if isinstance(self.axis, (list, tuple)) else [self.axis]
            for ii in axis:
                bb_shape[ii] = input_shape[ii]
        self.bb = self.add_weight(name="bias", shape=bb_shape, initializer=self.initializer, trainable=True)
        super(BiasLayer, self).build(input_shape)

    def call(self, inputs, **kwargs):
        return inputs + self.bb

    def get_config(self):
        config = super(BiasLayer, self).get_config()
        config.update({"axis": self.axis})  # Not saving initializer in config
        return config

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        if backend.image_data_format() != "channels_last" and self.axis == 1:
            weights = [np.squeeze(ii) for ii in weights]
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        if backend.image_data_format() != "channels_last" and self.axis == 1:
            weights = [np.reshape(ii, self.bb.shape) for ii in weights]
        return self.set_weights(weights)


def attention_mlp_block(inputs, embed_dim, num_heads=1, mlp_ratio=3, attention_type=None, drop_rate=0, mlp_activation="gelu", dropout=0, name=""):
    # print(f">>>> {drop_rate = }")
    nn_0 = inputs[:, :1] if attention_type == "class" else inputs
    nn_1 = layers.LayerNormalization(epsilon=BATCH_NORM_EPSILON, axis=-1, name=name + "LN")(inputs)
    key_dim = embed_dim // num_heads

    if attention_type == "outlook":
        nn_1 = outlook_attention(nn_1, embed_dim, num_heads=num_heads, name=name + "attn_")
    elif attention_type == "outlook_simple":
        nn_1 = outlook_attention_simple(nn_1, embed_dim, num_heads=num_heads, name=name + "attn_")
    elif attention_type == "class":
        # nn_1 = class_attention(nn_1, embed_dim, num_heads=num_heads, name=name + "attn_")
        query = nn_1[:, :1, :]
        nn_1 = layers.MultiHeadAttention(num_heads=num_heads, key_dim=key_dim, output_shape=embed_dim, use_bias=False, name=name + "attn_mhsa")(query, nn_1)
        nn_1 = BiasLayer(name=name + "attn_bias")(nn_1)  # bias for output dense
    elif attention_type == "mhsa":
        # nn_1 = multi_head_self_attention(nn_1, num_heads=num_heads, key_dim=key_dim, out_shape=embed_dim, out_weight=True, out_bias=True, name=name + "attn_")
        nn_1 = layers.MultiHeadAttention(num_heads=num_heads, key_dim=key_dim, output_shape=embed_dim, use_bias=False, name=name + "attn_mhsa")(nn_1, nn_1)
        nn_1 = BiasLayer(name=name + "attn_bias")(nn_1)  # bias for output dense

    if drop_rate > 0:
        nn_1 = layers.Dropout(drop_rate, noise_shape=(None, 1, 1, 1), name=name + "drop_1")(nn_1)
    nn_1 = layers.Add()([nn_0, nn_1])

    """ MLP """
    nn_2 = layers.LayerNormalization(epsilon=BATCH_NORM_EPSILON, axis=-1, name=name + "mlp_LN")(nn_1)
    nn_2 = layers.Dense(embed_dim * mlp_ratio, name=name + "mlp_dense_1")(nn_2)
    # gelu with approximate=False using `erf` leads to GPU memory leak...
    # nn_2 = layers.Activation("gelu", name=name + "mlp_" + mlp_activation)(nn_2)
    nn_2 = activation_by_name(nn_2, mlp_activation, name=name + mlp_activation)
    nn_2 = layers.Dense(embed_dim, name=name + "mlp_dense_2")(nn_2)
    if dropout > 0:
        nn_2 = layers.Dropout(dropout)(nn_2)

    if drop_rate > 0:
        nn_2 = layers.Dropout(drop_rate, noise_shape=(None, 1, 1, 1), name=name + "drop_2")(nn_2)
    out = layers.Add(name=name + "output")([nn_1, nn_2])

    if attention_type == "class":
        out = functional.concat([out, inputs[:, 1:]], axis=1)
    return out


@register_keras_serializable(package="volo")
class PositionalEmbedding(layers.Layer):
    def __init__(self, input_height=-1, **kwargs):
        super().__init__(**kwargs)
        self.pp_init = initializers.TruncatedNormal(stddev=0.2)
        self.input_height = input_height

    def build(self, input_shape):
        self.positional_embedding = self.add_weight(name="positional_embedding", shape=(1, *input_shape[1:]), initializer=self.pp_init, trainable=True)
        super().build(input_shape)

        if len(input_shape) == 3:
            # height and width in input_shape reshaped as one, like [None, 16 * 16 + 1, 32]
            self.is_fused_height_width = True
            self.height = self.input_height if self.input_height > 0 else int(float(input_shape[1]) ** 0.5)
            self.width = input_shape[1] // self.height
        else:
            self.is_fused_height_width = False
            self.height, self.width = input_shape[1:3]

    def call(self, inputs, **kwargs):
        return inputs + self.positional_embedding

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"input_height": self.input_height})
        return base_config

    def load_resized_weights(self, source_layer, method="bilinear"):
        # For input 224 --> [1, 14, 14, 384], convert to 384 --> [1, 24, 24, 384]
        if isinstance(source_layer, dict):
            source_pp = list(source_layer.values())[0]  # weights
        else:
            source_pp = source_layer.positional_embedding  # layer

        source_pp = np.array(source_pp).astype("float32")
        if self.is_fused_height_width:
            hh = ww = int(float(source_pp.shape[1]) ** 0.5)  # assume source weights are all square shape
            ss = source_pp[:, -hh * ww :]  # If has cls_token
            ss = ss.reshape([1, hh, ww, -1])

            tt = backend.numpy_image_resize(ss, target_shape=[self.height, self.width], method=method)
            tt = np.reshape(tt, [1, self.height * self.width, -1])
            tt = np.concatenate([source_pp[:, : -hh * ww], tt], axis=1)  # If has cls_token
        else:
            tt = backend.numpy_image_resize(source_pp, target_shape=[self.height, self.width], method=method)
        # functional.assign(self.positional_embedding, tt)  # For TF it's `parameter.assign(data)`, for Torch `parameter.data = torch.tensor(data)`
        self.set_weights([tt])

    def show_pos_emb(self, rows=16, base_size=1):
        import matplotlib.pyplot as plt

        ss = self.positional_embedding[0]
        cols = int(math.ceil(ss.shape[-1] / rows))
        fig, axes = plt.subplots(rows, cols, figsize=(base_size * cols, base_size * rows))
        for id, ax in enumerate(axes.flatten()):
            ax.imshow(ss[:, :, id])
            ax.set_axis_off()
        fig.tight_layout()
        return fig


@register_keras_serializable(package="volo")
class ClassToken(layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.token_init = initializers.TruncatedNormal(stddev=0.2)

    def build(self, input_shape):
        self.class_tokens = self.add_weight(name="tokens", shape=(1, 1, input_shape[-1]), initializer=self.token_init, trainable=True)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        if backend.is_torch_backend:
            class_tokens = self.class_tokens.expand(inputs.shape[0], -1, -1)
        else:
            class_tokens = functional.repeat(self.class_tokens, functional.shape(inputs)[0], axis=0)
        return functional.concat([class_tokens, inputs], axis=1)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[1] + 1, input_shape[2])


@register_keras_serializable(package="volo")
class MixupToken(layers.Layer):
    def __init__(self, scale=2, beta=1.0, **kwargs):
        super().__init__(**kwargs)
        self.scale, self.beta = scale, beta

    def call(self, inputs, training=None, **kwargs):
        height, width = functional.shape(inputs)[1], functional.shape(inputs)[2]
        # tf.print("training:", training)
        def _call_train():
            return functional.stack(self.rand_bbox(height, width))

        def _call_test():
            return functional.cast(functional.stack([0, 0, 0, 0]), "int32")  # No mixup area for test

        return backend.in_train_phase(_call_train, _call_test, training=training)

    def sample_beta_distribution(self):
        import tensorflow as tf

        gamma_1_sample = tf.random.gamma(shape=[], alpha=self.beta)
        gamma_2_sample = tf.random.gamma(shape=[], alpha=self.beta)
        return gamma_1_sample / (gamma_1_sample + gamma_2_sample)

    def rand_bbox(self, height, width):
        import tensorflow as tf

        random_lam = functional.cast(self.sample_beta_distribution(), self.compute_dtype)
        cut_rate = functional.sqrt(1.0 - random_lam)
        s_height, s_width = height // self.scale, width // self.scale

        right_pos = tf.random.uniform(shape=[], minval=0, maxval=s_width, dtype="int32")
        bottom_pos = tf.random.uniform(shape=[], minval=0, maxval=s_height, dtype="int32")
        left_pos = right_pos - functional.cast(functional.cast(s_width, cut_rate.dtype) * cut_rate, "int32") // 2
        top_pos = bottom_pos - functional.cast(functional.cast(s_height, cut_rate.dtype) * cut_rate, "int32") // 2
        left_pos, top_pos = functional.maximum(left_pos, 0), functional.maximum(top_pos, 0)

        return left_pos, top_pos, right_pos, bottom_pos

    def do_mixup_token(self, inputs, bbox):
        left, top, right, bottom = bbox
        sub_ww = inputs[:, :, left:right]
        mix_sub = functional.concat([sub_ww[:, :top], sub_ww[::-1, top:bottom], sub_ww[:, bottom:]], axis=1)
        output = functional.concat([inputs[:, :, :left], mix_sub, inputs[:, :, right:]], axis=2)
        output.set_shape(inputs.shape)
        return output

    def get_config(self):
        config = super().get_config()
        config.update({"scale": self.scale, "beta": self.beta})
        return config


def patch_stem(inputs, hidden_dim=64, stem_width=384, patch_size=8, strides=2, activation="relu", name=""):
    nn = conv2d_no_bias(inputs, hidden_dim, 7, strides=strides, padding="same", name=name + "1_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "1_")
    nn = conv2d_no_bias(nn, hidden_dim, 3, strides=1, padding="same", name=name + "2_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "2_")
    nn = conv2d_no_bias(nn, hidden_dim, 3, strides=1, padding="same", name=name + "3_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "3_")

    patch_step = patch_size // strides
    return conv2d_no_bias(nn, stem_width, patch_step, strides=patch_step, use_bias=True, name=name + "patch_")


def VOLO(
    num_blocks,
    embed_dims,
    num_heads,
    mlp_ratios,
    stem_hidden_dim=64,
    patch_size=8,
    mlp_activation="gelu",
    input_shape=(224, 224, 3),
    num_classes=1000,
    drop_connect_rate=0,
    classfiers=2,
    mix_token=False,
    token_classifier_top=False,
    mean_classifier_top=False,
    token_label_top=False,
    first_attn_type="outlook",
    pretrained="imagenet",
    model_name="VOLO",
    kwargs=None,
):
    inputs = layers.Input(input_shape)

    """ forward_embeddings """
    nn = patch_stem(inputs, hidden_dim=stem_hidden_dim, stem_width=embed_dims[0], patch_size=patch_size, strides=2, name="stem_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last

    if mix_token:
        scale = 2
        mixup_token = MixupToken(scale=scale)
        bbox = mixup_token(nn)
        nn = mixup_token.do_mixup_token(nn, bbox * scale)

    outlook_attentions = [True, False]
    downsamples = [True, False]

    """ forward_tokens """
    total_blocks = sum(num_blocks)
    global_block_id = 0

    # Outlook attentions
    num_block, embed_dim, num_head, mlp_ratio = num_blocks[0], embed_dims[0], num_heads[0], mlp_ratios[0]
    for ii in range(num_block):
        name = "outlook_block{}_".format(ii)
        block_drop_rate = drop_connect_rate * global_block_id / total_blocks
        nn = attention_mlp_block(nn, embed_dim, num_head, mlp_ratio, first_attn_type, block_drop_rate, mlp_activation, name=name)
        global_block_id += 1

    # downsample
    nn = layers.Conv2D(embed_dim * 2, kernel_size=2, strides=2, name="downsample_conv")(nn)
    # PositionalEmbedding
    nn = PositionalEmbedding(name="positional_embedding")(nn)

    # MHSA attentions
    num_block, embed_dim, num_head, mlp_ratio = num_blocks[1], embed_dims[1], num_heads[1], mlp_ratios[1]
    for ii in range(num_block):
        name = "MHSA_block{}_".format(ii)
        block_drop_rate = drop_connect_rate * global_block_id / total_blocks
        nn = attention_mlp_block(nn, embed_dim, num_head, mlp_ratio, "mhsa", block_drop_rate, mlp_activation, name=name)
        global_block_id += 1

    if num_classes == 0:
        model = models.Model(inputs, nn, name=model_name)
        reload_model_weights(model, PRETRAINED_DICT, "volo", pretrained, PositionalEmbedding)
        return model

    _, height, width, channel = nn.shape
    nn = functional.reshape(nn, (-1, height * width, channel))

    """ forward_cls """
    nn = ClassToken(name="class_token")(nn)

    embed_dim, num_head, mlp_ratio = embed_dims[-1], num_heads[-1], mlp_ratios[-1]
    for id in range(classfiers):
        name = "classfiers{}_".format(id)
        nn = attention_mlp_block(nn, embed_dim, num_head, mlp_ratio, "class", mlp_activation=mlp_activation, name=name)
    nn = layers.LayerNormalization(epsilon=BATCH_NORM_EPSILON, axis=-1, name="pre_out_LN")(nn)

    if token_label_top:
        # Training with label token
        nn_cls = layers.Dense(num_classes, dtype="float32", name="token_head")(nn[:, 0])
        nn_aux = layers.Dense(num_classes, dtype="float32", name="aux_head")(nn[:, 1:])

        if mix_token:
            nn_aux = functional.reshape(nn_aux, (-1, height, width, num_classes))
            nn_aux = mixup_token.do_mixup_token(nn_aux, bbox)
            nn_aux = layers.Reshape((height * width, num_classes), dtype="float32", name="aux")(nn_aux)

            left, top, right, bottom = bbox
            lam = 1 - ((right - left) * (bottom - top) / nn_aux.shape[1])
            lam_repeat = functional.expand_dims(functional.repeat(lam, functional.shape(inputs)[0], axis=0), 1)
            nn_cls = layers.Concatenate(axis=-1, dtype="float32", name="class")([nn_cls, lam_repeat])

        # nn_lam = layers.Lambda(lambda ii: functional.cast(functional.stack(ii), "float32"))([left_pos, top_pos, right_pos, bottom_pos])
        nn = [nn_cls, nn_aux]
    elif mean_classifier_top:
        # Return mean of all tokens
        nn = layers.GlobalAveragePooling1D(name="avg_pool")(nn)
        nn = layers.Dense(num_classes, dtype="float32", name="token_head")(nn)
    elif token_classifier_top:
        # Return dense classifier using only first token
        nn = layers.Dense(num_classes, dtype="float32", name="token_head")(nn[:, 0])
    else:
        # Return token dense for evaluation
        nn_cls = layers.Dense(num_classes, dtype="float32", name="token_head")(nn[:, 0])
        nn_aux = layers.Dense(num_classes, dtype="float32", name="aux_head")(nn[:, 1:])
        nn = layers.Add()([nn_cls, functional.reduce_max(nn_aux, 1) * 0.5])

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "volo", pretrained, PositionalEmbedding)
    return model


@register_model
def VOLO_d1(input_shape=(224, 224, 3), num_classes=1000, pretrained="imagenet", **kwargs):
    num_blocks = [4, 14]
    embed_dims = [192, 384]
    num_heads = [6, 12]
    mlp_ratios = [3, 3]
    stem_hidden_dim = 64
    return VOLO(**locals(), model_name="volo_d1", **kwargs)


@register_model
def VOLO_d2(input_shape=(224, 224, 3), num_classes=1000, pretrained="imagenet", **kwargs):
    num_blocks = [6, 18]
    embed_dims = [256, 512]
    num_heads = [8, 16]
    mlp_ratios = [3, 3]
    stem_hidden_dim = 64
    return VOLO(**locals(), model_name="volo_d2", **kwargs)


@register_model
def VOLO_d3(input_shape=(224, 224, 3), num_classes=1000, pretrained="imagenet", **kwargs):
    num_blocks = [8, 28]
    embed_dims = [256, 512]
    num_heads = [8, 16]
    mlp_ratios = [3, 3]
    stem_hidden_dim = 64
    return VOLO(**locals(), model_name="volo_d3", **kwargs)


@register_model
def VOLO_d4(input_shape=(224, 224, 3), num_classes=1000, pretrained="imagenet", **kwargs):
    num_blocks = [8, 28]
    embed_dims = [384, 768]
    num_heads = [12, 16]
    mlp_ratios = [3, 3]
    stem_hidden_dim = 64
    return VOLO(**locals(), model_name="volo_d4", **kwargs)


@register_model
def VOLO_d5(input_shape=(224, 224, 3), num_classes=1000, pretrained="imagenet", **kwargs):
    num_blocks = [12, 36]
    embed_dims = [384, 768]
    num_heads = [12, 16]
    mlp_ratios = [4, 4]
    stem_hidden_dim = 128
    return VOLO(**locals(), model_name="volo_d5", **kwargs)



from keras_cv_attention_models.maxvit.maxvit import MaxViT, MaxViT_Tiny, MaxViT_Small, MaxViT_Base, MaxViT_Large, MaxViT_XLarge

__head_doc__ = """
Keras implementation of [Github google-research/maxvit](https://github.com/google-research/maxvit).
Paper [PDF 2204.01697 MaxViT: Multi-Axis Vision Transformer](https://arxiv.org/pdf/2204.01697.pdf).
"""

__tail_doc__ = """  strides: int or list of int, for strides in the first block of each stack.
  expansion: filter expansion in each block. The larger the wider.
  se_ratio: value in `(0, 1)`, where `0` means not using `se_module`. Should be a `number` or `list`, indicates `se_ratio` for each stack.
      Each element can also be a `number` or `list`, indicates `se_ratio` for each block.
  head_dimension: heads number for transformer block.
  window_ratio: window_size ratio, window_size = [input_shape[0] // window_ratio, input_shape[1] // window_ratio].
  output_filter: additional `Dense + tanh` block before ouput block. Default -1 for `out_channels[-1]`, 0 to disable.
  use_torch_mode: whether use torch model or not.
      True: use_torch_padding = True, epsilon = 1e-5, momentum = 0.9
      False: use_torch_padding = False, epsilon = 0.001, momentum = 0.99
  layer_scale: layer scale init value. `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf). Default -1.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
      Actual `num_classes` for `"imagenet21k"` models is `21843`.
  activation: activation used in whole model, default "gelu/app" means `tf.nn.gelu(approximate=True)`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: One of `[None, "imagenet", "imagenet21k", "imagenet21k-ft1k"]`.
  **kwargs: other parameters if available.

Returns:
  A `keras.Model` instance.
"""

MaxViT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  stem_width: output dimension for stem block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                           | Params | FLOPs  | Input | Top1 Acc |
  | ------------------------------- | ------ | ------ | ----- | -------- |
  | MaxViT_Tiny                     | 31M    | 5.6G   | 224   | 83.62    |
  | MaxViT_Tiny                     | 31M    | 17.7G  | 384   | 85.24    |
  | MaxViT_Tiny                     | 31M    | 33.7G  | 512   | 85.72    |
  | MaxViT_Small                    | 69M    | 11.7G  | 224   | 84.45    |
  | MaxViT_Small                    | 69M    | 36.1G  | 384   | 85.74    |
  | MaxViT_Small                    | 69M    | 67.6G  | 512   | 86.19    |
  | MaxViT_Base                     | 119M   | 24.2G  | 224   | 84.95    |
  | - imagenet21k                   | 135M   | 24.2G  | 224   |          |
  | MaxViT_Base                     | 119M   | 74.2G  | 384   | 86.34    |
  | - imagenet21k-ft1k              | 119M   | 74.2G  | 384   | 88.24    |
  | MaxViT_Base                     | 119M   | 138.5G | 512   | 86.66    |
  | - imagenet21k-ft1k              | 119M   | 138.5G | 512   | 88.38    |
  | MaxViT_Large                    | 212M   | 43.9G  | 224   | 85.17    |
  | - imagenet21k                   | 233M   | 43.9G  | 224   |          |
  | MaxViT_Large                    | 212M   | 133.1G | 384   | 86.40    |
  | - imagenet21k-ft1k              | 212M   | 133.1G | 384   | 88.32    |
  | MaxViT_Large                    | 212M   | 245.4G | 512   | 86.70    |
  | - imagenet21k-ft1k              | 212M   | 245.4G | 512   | 88.46    |
  | MaxViT_XLarge, imagenet21k      | 507M   | 97.7G  | 224   |          |
  | MaxViT_XLarge, imagenet21k-ft1k | 475M   | 293.7G | 384   | 88.51    |
  | MaxViT_XLarge, imagenet21k-ft1k | 475M   | 535.2G | 512   | 88.70    |
"""

MaxViT_Tiny.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

MaxViT_Small.__doc__ = MaxViT_Tiny.__doc__
MaxViT_Base.__doc__ = MaxViT_Tiny.__doc__
MaxViT_Large.__doc__ = MaxViT_Tiny.__doc__
MaxViT_XLarge.__doc__ = MaxViT_Tiny.__doc__



from keras_cv_attention_models.fastervit.fastervit import (
    FasterViT,
    FasterViT0,
    FasterViT1,
    FasterViT2,
    FasterViT3,
    FasterViT4,
    FasterViT5,
    FasterViT6,
    switch_to_deploy,
)

__head_doc__ = """
Keras implementation of [Github NVlabs/FasterViT](https://github.com/NVlabs/FasterViT).
Paper [PDF 2306.06189 FasterViT: Fast Vision Transformers with Hierarchical Attention](https://arxiv.org/pdf/2306.06189.pdf).
"""

__tail_doc__ = """  window_ratios: window split ratio. It's mainly for the 3rd stack, that `window_size = (height // window_ratio, width // window_ratio)`.
      `1` means not using window partition, while `window_size == (height, width)`.
  carrier_token_size: int value indicates carrier token size for the 3rd stack.
  pos_scale: If pretrained weights are from different input_shape or window_size, pos_scale is previous actually using window_size.
  use_propagation: boolean value if using `do_propagation` block at the end of the 3rd stack.
  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: one of None or "imagenet".
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

FasterViT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  num_heads: num heads for each stack.
  stem_hidden_dim: hidden dimension for the 1st stem `Conv2D`.
  embed_dim: basic hidden dims, expand * 2 for each stack.
  mlp_ratio: expand ratio for mlp blocks hidden channel.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model      | Params   | FLOPs   | Input | Top1 Acc |
  | ---------- | -------- | ------- | ----- | -------- |
  | FasterViT0 | 31.40M   | 3.51G   | 224   | 82.1     |
  | FasterViT1 | 53.37M   | 5.52G   | 224   | 83.2     |
  | FasterViT2 | 75.92M   | 9.00G   | 224   | 84.2     |
  | FasterViT3 | 159.55M  | 18.75G  | 224   | 84.9     |
  | FasterViT4 | 351.12M  | 41.57G  | 224   | 85.4     |
  | FasterViT5 | 957.52M  | 114.08G | 224   | 85.6     |
  | FasterViT6 | 1360.33M | 144.13G | 224   | 85.8     |
"""

FasterViT0.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

FasterViT1.__doc__ = FasterViT0.__doc__
FasterViT2.__doc__ = FasterViT0.__doc__
FasterViT3.__doc__ = FasterViT0.__doc__
FasterViT4.__doc__ = FasterViT0.__doc__
FasterViT5.__doc__ = FasterViT0.__doc__
FasterViT6.__doc__ = FasterViT0.__doc__



from keras_cv_attention_models.beit.beit import Beit, keras_model_load_weights_from_pytorch_model
from keras_cv_attention_models.models import register_model


def FlexiViT(
    attn_qv_bias=False,
    attn_qkv_bias=True,
    use_abs_pos_emb=True,
    use_abs_pos_emb_on_cls_token=False,  # no_embed_class in timm
    layer_scale=0,
    use_mean_pooling_head=False,
    model_name="flexivit",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    patch_size = kwargs.pop("patch_size", 16)
    force_reload_mismatch = patch_size != 16  # If patch_size not 16, force reload pos_emb and stem_conv weights
    return Beit(**locals(), **kwargs)


@register_model
def FlexiViTSmall(input_shape=(240, 240, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 384
    depth = 12
    num_heads = 6
    return FlexiViT(**locals(), model_name="flexivit_small", **kwargs)


@register_model
def FlexiViTBase(input_shape=(240, 240, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 768
    depth = 12
    num_heads = 12
    return FlexiViT(**locals(), model_name="flexivit_base", **kwargs)


@register_model
def FlexiViTLarge(input_shape=(240, 240, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 1024
    depth = 24
    num_heads = 16
    return FlexiViT(**locals(), model_name="flexivit_large", **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    BiasLayer,
    ChannelAffine,
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    add_pre_post_process,
)
from keras_cv_attention_models import model_surgery
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.coco import eval_func, anchors_func

PRETRAINED_DICT = {
    "yolor_csp": {"coco": "ed0aa82a07c4e65e9cd3d2e6ad2d0548"},
    "yolor_csp_x": {"coco": "615125ce1cd1c855f8045bf079456598"},
    "yolor_p6": {"coco": "059c6d0dd8ca869f843081b13f88f7f4"},
    "yolor_w6": {"coco": "a3dc1e70c5064aebfd8b52609e6ee704"},
    "yolor_e6": {"coco": "556263cf6aeea5b628c1814cd126eb21"},
    "yolor_d6": {"coco": "a55469feef931a07b419c3e1be639725"},
}


""" CSPDarknet backbone """
BATCH_NORM_EPSILON = 1e-4
BATCH_NORM_MOMENTUM = 0.97


def conv_dw_pw_block(inputs, filters, kernel_size=1, strides=1, use_depthwise_conv=False, activation="swish", name=""):
    nn = inputs
    if use_depthwise_conv:
        nn = depthwise_conv2d_no_bias(nn, kernel_size, strides, padding="same", name=name)
        nn = batchnorm_with_activation(nn, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name + "dw_")
        kernel_size, strides = 1, 1
    nn = conv2d_no_bias(nn, filters, kernel_size, strides, padding="same", name=name)
    nn = batchnorm_with_activation(nn, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name)
    return nn


def csp_block(inputs, expansion=0.5, use_shortcut=True, activation="swish", name=""):
    input_channels = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    nn = conv_dw_pw_block(inputs, int(input_channels * expansion), activation=activation, name=name + "1_")
    nn = conv_dw_pw_block(nn, input_channels, kernel_size=3, strides=1, activation=activation, name=name + "2_")
    if use_shortcut:
        nn = layers.Add()([inputs, nn])
    return nn


def csp_stack(
    inputs, depth, out_channels=-1, expansion=0.5, use_shortcut=True, use_pre=False, use_post=True, use_shortcut_bn=True, activation="swish", name=""
):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    out_channels = inputs.shape[channel_axis] if out_channels == -1 else out_channels
    hidden_channels = int(out_channels * expansion)
    if use_pre:
        inputs = conv_dw_pw_block(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "pre_")
    if use_shortcut_bn:
        short = conv_dw_pw_block(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "short_")
    else:
        short = conv2d_no_bias(inputs, hidden_channels, 1, name=name + "short_")

    deep = inputs if use_pre else conv_dw_pw_block(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "deep_pre_")
    for id in range(depth):
        block_name = name + "block{}_".format(id + 1)
        deep = csp_block(deep, 1, use_shortcut=use_shortcut, activation=activation, name=block_name)
    if use_post:
        deep = conv_dw_pw_block(deep, hidden_channels, kernel_size=1, activation=activation, name=name + "deep_post_")

    out = functional.concat([deep, short], axis=channel_axis)
    if not use_shortcut_bn:
        out = batchnorm_with_activation(out, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name + "concat_")

    out = conv_dw_pw_block(out, out_channels, kernel_size=1, activation=activation, name=name + "output_")
    return out


def res_spatial_pyramid_pooling(inputs, depth, expansion=0.5, pool_sizes=(5, 9, 13), use_shortcut_bn=True, activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channels = inputs.shape[channel_axis]
    hidden_channels = int(input_channels * expansion)
    if use_shortcut_bn:
        short = conv_dw_pw_block(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "short_")
    else:
        short = conv2d_no_bias(inputs, hidden_channels, 1, name=name + "short_")

    deep = conv_dw_pw_block(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "pre_1_")
    deep = conv_dw_pw_block(deep, hidden_channels, kernel_size=3, activation=activation, name=name + "pre_2_")
    deep = conv_dw_pw_block(deep, hidden_channels, kernel_size=1, activation=activation, name=name + "pre_3_")
    pp = [layers.MaxPool2D(pool_size=ii, strides=1, padding="same")(deep) for ii in pool_sizes]
    deep = functional.concat([deep, *pp][::-1], axis=channel_axis)  # yolor_csp.cfg, SSP concat layers=-1,-3,-5,-6
    for id in range(depth - 1):  # First one is `pre`
        deep = conv_dw_pw_block(deep, hidden_channels, kernel_size=1, activation=activation, name=name + "post_{}_".format(id * 2 + 1))
        deep = conv_dw_pw_block(deep, hidden_channels, kernel_size=3, activation=activation, name=name + "post_{}_".format(id * 2 + 2))

    out = functional.concat([deep, short], axis=channel_axis)
    if not use_shortcut_bn:
        out = batchnorm_with_activation(out, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name + "concat_")
    out = conv_dw_pw_block(out, hidden_channels, kernel_size=1, activation=activation, name=name + "output_")
    return out


def focus_stem(inputs, filters, kernel_size=3, strides=1, padding="valid", activation="swish", name=""):
    is_channels_last = image_data_format() == "channels_last"
    channel_axis = -1 if is_channels_last else 1
    if padding.lower() == "same":  # Handling odd input_shape
        inputs = functional.pad(inputs, [[0, 0], [0, 1], [0, 1], [0, 0]] if is_channels_last else [[0, 0], [0, 0], [0, 1], [0, 1]])
        patch_top_left = inputs[:, :-1:2, :-1:2] if is_channels_last else inputs[:, :, :-1:2, :-1:2]
        patch_top_right = inputs[:, :-1:2, 1::2] if is_channels_last else inputs[:, :, :-1:2, 1::2]
        patch_bottom_left = inputs[:, 1::2, :-1:2] if image_data_format() == "channels_last" else inputs[:, :, 1::2, :-1:2]
        patch_bottom_right = inputs[:, 1::2, 1::2] if image_data_format() == "channels_last" else inputs[:, :, 1::2, 1::2]
    else:
        patch_top_left = inputs[:, ::2, ::2] if is_channels_last else inputs[:, :, ::2, ::2]
        patch_top_right = inputs[:, ::2, 1::2] if is_channels_last else inputs[:, :, ::2, 1::2]
        patch_bottom_left = inputs[:, 1::2, ::2] if is_channels_last else inputs[:, :, 1::2, ::2]
        patch_bottom_right = inputs[:, 1::2, 1::2] if is_channels_last else inputs[:, :, 1::2, 1::2]
    nn = functional.concat([patch_top_left, patch_bottom_left, patch_top_right, patch_bottom_right], axis=channel_axis)
    nn = conv_dw_pw_block(nn, filters, kernel_size=kernel_size, strides=strides, activation=activation, name=name)
    return nn


def csp_conv_downsample(inputs, filters, strides=2, activation="swish", name=""):
    # DownC: https://github.com/WongKinYiu/yolor/blob/paper/models/common.py#L584
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    max_down = layers.MaxPool2D(strides, strides=strides, padding="same")(inputs)
    max_down = conv_dw_pw_block(max_down, filters // 2, activation=activation, name=name + "max_down_")
    conv_down = conv_dw_pw_block(inputs, inputs.shape[channel_axis], activation=activation, name=name + "conv_down_1_")
    conv_down = conv_dw_pw_block(conv_down, filters // 2, kernel_size=3, strides=strides, activation=activation, name=name + "conv_down_2_")
    return functional.concat([conv_down, max_down], axis=channel_axis)


def CSPDarknet(
    depthes=[2, 8, 8, 4],
    channels=[128, 256, 512, 1024],
    stem_width=-1,  # -1 means using channels[0] // 2
    use_focus_stem=False,
    ssp_depth=2,
    out_features=[-3, -2, -1],
    use_csp_downsample=False,
    use_shortcut_bn=True,
    use_pre=False,
    use_post=True,
    input_shape=(512, 512, 3),
    activation="swish",
    model_name="",
):
    # base_channels = int(width_mul * 64)
    inputs = layers.Input(backend.align_input_shape_by_image_data_format(input_shape))

    """ Stem """
    stem_width = stem_width if stem_width > 0 else channels[0] // 2
    if use_focus_stem:
        nn = focus_stem(inputs, stem_width, activation=activation, name="stem_")
    else:
        nn = conv_dw_pw_block(inputs, 32, kernel_size=3, strides=1, activation=activation, name="stem_1_")  # Fixed as 32
        nn = conv_dw_pw_block(nn, stem_width, kernel_size=3, strides=2, activation=activation, name="stem_2_")
        nn = csp_block(nn, expansion=0.5, activation=activation, name="stem_3_")
    features = [nn]

    """ dark blocks """
    # depthes = [max(round(depth_mul * ii), 1) for ii in [2, 8, 8, 4]]  # YOLOR_CSP depth
    # channels = [base_channels * 2, base_channels * 4, base_channels * 8, base_channels * 16]
    # use_spps = [False] * (len(depthes) - 1) + [True]
    for id, (channel, depth) in enumerate(zip(channels, depthes)):
        stack_name = "stack{}_".format(id + 1)
        if use_csp_downsample:
            nn = csp_conv_downsample(nn, channel, activation=activation, name=stack_name)
        else:
            nn = conv_dw_pw_block(nn, channel, 3, strides=2, activation=activation, name=stack_name + "downsample_")
        nn = csp_stack(nn, depth, use_pre=use_pre, use_post=use_post, use_shortcut_bn=use_shortcut_bn, activation=activation, name=stack_name)
        if id == len(depthes) - 1:
            # ssp_depth = max(round(depth_mul * 2), 1)
            nn = res_spatial_pyramid_pooling(nn, ssp_depth, use_shortcut_bn=use_shortcut_bn, activation=activation, name=stack_name + "spp_")
        features.append(nn)

    nn = [features[ii] for ii in out_features]
    model = models.Model(inputs, nn, name=model_name)
    return model


""" path aggregation fpn """


def upsample_merge(inputs, csp_depth, use_shortcut_bn=True, activation="swish", name=""):
    # print(f">>>> upsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    upsample = conv_dw_pw_block(inputs[-1], inputs[0].shape[channel_axis], activation=activation, name=name + "up_")

    # inputs[0] = layers.UpSampling2D(size=(2, 2), interpolation="nearest", name=name + "up")(fpn_out)
    size = functional.shape(inputs[0])[1:-1] if image_data_format() == "channels_last" else functional.shape(inputs[0])[2:]
    inputs[-1] = functional.resize(upsample, size, method="nearest")
    nn = functional.concat(inputs, axis=channel_axis)
    use_shortcut, use_pre, use_post = False, True, False
    nn = csp_stack(nn, csp_depth, nn.shape[channel_axis] // 2, 1.0, use_shortcut, use_pre, use_post, use_shortcut_bn, activation=activation, name=name)
    return nn


def downsample_merge(inputs, csp_depth, use_csp_downsample=False, use_shortcut_bn=True, activation="swish", name=""):
    # print(f">>>> downsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    if use_csp_downsample:
        inputs[0] = csp_conv_downsample(inputs[0], inputs[-1].shape[channel_axis], activation=activation, name=name)
    else:
        inputs[0] = conv_dw_pw_block(inputs[0], inputs[-1].shape[channel_axis], 3, 2, activation=activation, name=name + "down_")
    nn = functional.concat(inputs, axis=channel_axis)
    use_shortcut, use_pre, use_post = False, True, False
    nn = csp_stack(nn, csp_depth, nn.shape[channel_axis] // 2, 1.0, use_shortcut, use_pre, use_post, use_shortcut_bn, activation=activation, name=name)
    return nn


def path_aggregation_fpn(features, fpn_depth=2, use_csp_downsample=False, use_shortcut_bn=True, activation="swish", name=""):
    # p5 ---┬-----------------┬-> out2
    #       ↓ [up -> concat]  ↑ [down -> concat]
    # p4 -> p4p5 ------------> out1
    #       ↓ [up -> concat]  ↑ [down -> concat]
    # p3 -> p3p4p5 ------------┴-> out0
    # features: [p3, p4, p5]
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    upsamples = [features[-1]]
    p_name = "p{}_".format(len(features) + 2)
    # upsamples: [p5], features[:-1][::-1]: [p4, p3] -> [p5, p4p5, p3p4p5]
    for id, ii in enumerate(features[:-1][::-1]):
        cur_p_name = "p{}".format(len(features) + 1 - id)
        nn = conv_dw_pw_block(ii, ii.shape[channel_axis] // 2, kernel_size=1, activation=activation, name=name + cur_p_name + "_down_")
        p_name = cur_p_name + p_name
        nn = upsample_merge([nn, upsamples[-1]], fpn_depth, use_shortcut_bn, activation=activation, name=name + p_name)
        upsamples.append(nn)

    downsamples = [upsamples[-1]]
    # downsamples: [p3p4p5], upsamples[:-1][::-1]: [p4p5, p5] -> [p3p4p5, p3p4p5 + p4p5, p3p4p5 + p4p5 + p5]
    for id, ii in enumerate(upsamples[:-1][::-1]):
        cur_name = name + "c3n{}_".format(id + 3)
        nn = downsample_merge([downsamples[-1], ii], fpn_depth, use_csp_downsample, use_shortcut_bn, activation=activation, name=cur_name)
        downsamples.append(nn)
    return downsamples


""" YOLORHead """


def yolor_head_single(inputs, filters, num_classes=80, regression_len=4, num_anchors=3, use_object_scores=True, activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1

    nn = conv_dw_pw_block(inputs, filters, 3, activation=activation, name=name + "1_")
    nn = BiasLayer(initializer=initializers.truncated_normal(stddev=0.2), axis=channel_axis, name=name + "shift_channel")(nn)

    ouput_classes = num_classes + regression_len + (1 if use_object_scores else 0)  # num_anchors = 3, num_anchors * (80 + 5) = 255
    nn = layers.Conv2D(ouput_classes * num_anchors, kernel_size=1, name=name + "2_conv")(nn)
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)
    control_channels_layer = ChannelAffine(use_bias=False, axis=-1, name=name + "control_channel")
    control_channels_layer.ww_init = initializers.truncated_normal(stddev=0.2)
    nn = control_channels_layer(nn)
    # return nn
    return layers.Reshape([-1, ouput_classes], name=name + "output_reshape")(nn)


def yolor_head(inputs, num_classes=80, regression_len=4, num_anchors=1, use_object_scores=True, activation="swish", classifier_activation="sigmoid", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    outputs = []
    for id, input in enumerate(inputs):
        cur_name = name + "{}_".format(id + 1)
        filters = int(input.shape[channel_axis] * 2)
        out = yolor_head_single(input, filters, num_classes, regression_len, num_anchors, use_object_scores, activation=activation, name=cur_name)
        outputs.append(out)
    # return outputs
    outputs = functional.concat(outputs, axis=1)
    return activation_by_name(outputs, classifier_activation, name="classifier_")


""" YOLOR models """


def YOLOR(
    backbone=None,
    csp_depthes=[2, 8, 8, 4],
    csp_channels=[128, 256, 512, 1024],
    stem_width=-1,  # -1 means using csp_channels[0] // 2
    use_focus_stem=False,
    ssp_depth=2,
    csp_use_pre=False,
    csp_use_post=True,
    use_csp_downsample=False,
    use_shortcut_bn=True,
    fpn_depth=2,
    features_pick=[-3, -2, -1],
    regression_len=4,  # bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64
    anchors_mode="yolor",
    num_anchors="auto",  # "auto" means: anchors_mode=="anchor_free" -> 1, anchors_mode=="yolor" -> 3, else 9
    use_object_scores="auto",  # "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False
    input_shape=(640, 640, 3),
    num_classes=80,
    activation="swish",
    classifier_activation="sigmoid",
    freeze_backbone=False,
    pretrained=None,
    model_name="yolor",
    pyramid_levels_min=3,  # Init anchors for model prediction.
    anchor_scale="auto",  # Init anchors for model prediction. "auto" means 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4
    rescale_mode="raw01",  # For decode predictions, raw01 means input value in range [0, 1].
    kwargs=None,  # Not using, recieving parameter
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)

    if backbone is None:
        # Save line width...
        csp_kwargs = {"use_pre": csp_use_pre, "use_post": csp_use_post, "input_shape": input_shape, "activation": activation, "model_name": "darknet"}
        backbone = CSPDarknet(
            csp_depthes, csp_channels, stem_width, use_focus_stem, ssp_depth, features_pick, use_csp_downsample, use_shortcut_bn, **csp_kwargs
        )
        features = backbone.outputs
    else:
        if isinstance(features_pick[0], str):
            features = [backbone.get_layer(layer_name) for layer_name in features_pick]
        else:
            features = model_surgery.get_pyramide_feature_layers(backbone)
            features = [features[id] for id in features_pick]
        feature_names, features = model_surgery.align_pyramide_feature_output_by_image_data_format(features)
        print(">>>> features:", {ii: jj.shape for ii, jj in zip(feature_names, features)})

    backbone.trainable = False if freeze_backbone else True
    use_object_scores, num_anchors, anchor_scale = anchors_func.get_anchors_mode_parameters(anchors_mode, use_object_scores, num_anchors, anchor_scale)
    inputs = backbone.inputs[0]

    fpn_features = path_aggregation_fpn(features, fpn_depth, use_csp_downsample, use_shortcut_bn, activation=activation, name="pafpn_")
    outputs = yolor_head(fpn_features, num_classes, regression_len, num_anchors, use_object_scores, activation, classifier_activation, name="head_")
    outputs = layers.Activation("linear", dtype="float32", name="outputs_fp32")(outputs)
    model = models.Model(inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "yolor", pretrained)

    pyramid_levels = [pyramid_levels_min, pyramid_levels_min + len(features_pick) - 1]  # -> [3, 5]
    post_process = eval_func.DecodePredictions(
        backbone.input_shape[1:], pyramid_levels, anchors_mode, use_object_scores, anchor_scale, regression_len=regression_len
    )
    add_pre_post_process(model, rescale_mode=rescale_mode, post_process=post_process)
    return model


@register_model
def YOLOR_CSP(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_depthes = [2, 8, 8, 4]
    csp_channels = [128, 256, 512, 1024]
    fpn_depth = 2
    ssp_depth = 2
    return YOLOR(**locals(), model_name=kwargs.pop("model_name", "yolor_csp"), **kwargs)


@register_model
def YOLOR_CSPX(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_depthes = [3, 10, 10, 5]
    csp_channels = [160, 320, 640, 1280]
    fpn_depth = 3
    ssp_depth = 3
    return YOLOR(**locals(), model_name=kwargs.pop("model_name", "yolor_csp_x"), **kwargs)


@register_model
def YOLOR_P6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_depthes = [3, 7, 7, 3, 3]
    csp_channels = [128, 256, 384, 512, 640]
    features_pick = [-4, -3, -2, -1]
    fpn_depth = 3
    ssp_depth = 2
    use_focus_stem = True
    csp_use_post = False
    return YOLOR(**locals(), model_name=kwargs.pop("model_name", "yolor_p6"), **kwargs)


@register_model
def YOLOR_W6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_depthes = [3, 7, 7, 3, 3]
    csp_channels = [128, 256, 512, 768, 1024]
    features_pick = [-4, -3, -2, -1]
    fpn_depth = 3
    ssp_depth = 2
    use_focus_stem = True
    csp_use_post = False
    return YOLOR(**locals(), model_name=kwargs.pop("model_name", "yolor_w6"), **kwargs)


@register_model
def YOLOR_E6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_depthes = [3, 7, 7, 3, 3]
    csp_channels = [160, 320, 640, 960, 1280]
    features_pick = [-4, -3, -2, -1]
    fpn_depth = 3
    ssp_depth = 2
    use_focus_stem = True
    csp_use_post = False
    use_csp_downsample = True
    use_shortcut_bn = False
    return YOLOR(**locals(), model_name=kwargs.pop("model_name", "yolor_e6"), **kwargs)


@register_model
def YOLOR_D6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_depthes = [3, 15, 15, 7, 7]
    csp_channels = [160, 320, 640, 960, 1280]
    features_pick = [-4, -3, -2, -1]
    fpn_depth = 3
    ssp_depth = 2
    use_focus_stem = True
    csp_use_post = False
    use_csp_downsample = True
    use_shortcut_bn = False
    return YOLOR(**locals(), model_name=kwargs.pop("model_name", "yolor_d6"), **kwargs)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    conv2d_no_bias,
    group_norm,
    add_pre_post_process,
    qkv_to_multi_head_channels_last_format,
    scaled_dot_product_attention,
)
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.stable_diffusion.eval_func import RunPrediction

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {"unet": {"v1_5": "30b5e8755d2a04211980e608df14f133"}}


@backend.register_keras_serializable(package="kecam")
class SinusoidalTimeStepEmbedding(layers.Layer):
    def __init__(self, hidden_channels=320, max_period=10000, **kwargs):
        super().__init__(**kwargs)
        self.hidden_channels, self.max_period = hidden_channels, max_period

    def build(self, input_shape):
        # input_shape: [batch]
        half = self.hidden_channels // 2  # half the channels are sin and the other half is cos
        frequencies = np.exp(-np.log(self.max_period) * np.arange(0, half, dtype="float32") / half)
        positions = np.arange(self.max_period).astype("float32")
        embeddings = positions[:, None] * frequencies[None]
        embeddings = np.concatenate([np.cos(embeddings), np.sin(embeddings)], axis=-1).astype("float32")

        if hasattr(self, "register_buffer"):  # PyTorch
            self.register_buffer("embeddings", functional.convert_to_tensor(embeddings, dtype=self.compute_dtype), persistent=False)
        else:
            self.embeddings = functional.convert_to_tensor(embeddings, dtype=self.compute_dtype)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        return functional.embedding_lookup(self.embeddings, inputs)

    def compute_output_shape(self, input_shape):
        return [None, self.hidden_channels]

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"hidden_channels": self.hidden_channels, "max_period": self.max_period})
        return base_config


def cross_attention(inputs, condition=None, num_heads=4, head_dim=0, name=""):
    _, bb, cc = inputs.shape
    head_dim = head_dim if head_dim > 0 else cc // num_heads
    emded_dim = int(num_heads * head_dim)
    condition = inputs if condition is None else condition

    query = layers.Dense(emded_dim, use_bias=False, name=name and name + "query")(inputs)
    key = layers.Dense(emded_dim, use_bias=False, name=name and name + "key")(condition)
    value = layers.Dense(emded_dim, use_bias=False, name=name and name + "value")(condition)

    query, key, value = qkv_to_multi_head_channels_last_format(query, key, value, num_heads, data_format="channels_last")
    output_shape = [-1, -1 if bb is None else bb, cc]
    return scaled_dot_product_attention(query, key, value, output_shape, out_weight=True, out_bias=True, name=name)


def attention_mlp_block(inputs, condition=None, mlp_ratio=4, num_heads=4, head_dim=0, name=""):
    nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "attn_ln")(inputs)  # "channels_first" also using axis=-1
    nn = cross_attention(nn, condition=None, num_heads=num_heads, head_dim=head_dim, name=name + "attn_")
    attn_out = layers.Add(name=name + "attn_out")([inputs, nn])

    """ Attention with condition """
    if condition is not None:
        nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "cond_attn_ln")(attn_out)  # "channels_first" also using axis=-1
        nn = cross_attention(nn, condition=condition, num_heads=num_heads, head_dim=head_dim, name=name + "cond_attn_")
        attn_out = layers.Add(name=name + "cond_attn_out")([attn_out, nn])

    """ Feed forward """
    input_channels = inputs.shape[-1]
    nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "ffn_ln")(attn_out)  # "channels_first" also using axis=-1
    nn = layers.Dense(input_channels * mlp_ratio * 2, use_bias=True, name=name + "ffn_gate_dense")(nn)
    fft, fft_gate = functional.split(nn, 2, axis=-1)
    nn = fft * activation_by_name(fft_gate, activation="gelu", name=name + "gate_")

    nn = layers.Dense(input_channels, use_bias=True, name=name + "mlp.down_proj")(nn)
    nn = layers.Add(name=name + "mlp_output")([attn_out, nn])
    return nn


def spatial_transformer_block(inputs, condition=None, num_attention_block=1, mlp_ratio=4, num_heads=4, head_dim=0, name=""):
    input_channels = inputs.shape[-1 if backend.image_data_format() == "channels_last" else 1]
    nn = group_norm(inputs, epsilon=LAYER_NORM_EPSILON, name=name + "in_layers_")
    nn = conv2d_no_bias(nn, input_channels, kernel_size=1, use_bias=True, name=name + "in_layers_")

    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)
    pre_shape = functional.shape(nn) if None in nn.shape[1:] or -1 in nn.shape[1:] else [-1, *nn.shape[1:]]  # Could be dynamic shape, reshape back later
    nn = layers.Reshape([-1, nn.shape[-1]])(nn)

    for attention_block_id in range(num_attention_block):
        block_name = name + "{}_".format(attention_block_id + 1)
        nn = attention_mlp_block(nn, condition, mlp_ratio=mlp_ratio, num_heads=num_heads, head_dim=head_dim, name=block_name)
    nn = functional.reshape(nn, pre_shape)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)
    nn = conv2d_no_bias(nn, input_channels, kernel_size=1, use_bias=True, name=name + "out_layers_")
    return layers.Add(name=name + "output")([inputs, nn])


def res_block(inputs, time_embedding=None, labels_embedding=None, out_channels=-1, epsilon=1e-5, activation="swish", dropout=0, name=""):
    input_channels = inputs.shape[-1 if backend.image_data_format() == "channels_last" else 1]
    out_channels = out_channels if out_channels > 0 else input_channels
    if input_channels == out_channels:
        short = inputs
    else:
        short = conv2d_no_bias(inputs, out_channels, kernel_size=1, use_bias=True, name=name + "short_")

    nn = group_norm(inputs, epsilon=epsilon, name=name + "in_layers_")
    nn = activation_by_name(nn, activation=activation, name=name + "in_layers_")
    nn = conv2d_no_bias(nn, out_channels, kernel_size=3, use_bias=True, padding="SAME", name=name + "in_layers_")

    if time_embedding is not None:
        emb = activation_by_name(time_embedding, activation=activation, name=name + "emb_layers_")
        emb = layers.Dense(out_channels, name=name + "emb_layers_dense")(emb)
        emb = emb[:, None, None, :] if backend.image_data_format() == "channels_last" else emb[:, :, None, None]
        nn = nn + emb

    if labels_embedding is not None:
        emb = activation_by_name(labels_embedding, activation=activation, name=name + "labels_emb_layers_")
        emb = layers.Dense(out_channels, name=name + "labels_emb_layers_dense")(emb)
        emb = emb[:, None, None, :] if backend.image_data_format() == "channels_last" else emb[:, :, None, None]
        nn = nn + emb

    nn = group_norm(nn, epsilon=epsilon, name=name + "out_layers_")
    nn = activation_by_name(nn, activation=activation, name=name + "out_layers_")
    nn = conv2d_no_bias(nn, out_channels, kernel_size=3, use_bias=True, padding="SAME", name=name + "out_layers_")
    nn = layers.Dropout(dropout=dropout)(nn) if dropout > 0 else nn
    return layers.Add(name=name + "out")([nn, short])


@register_model
def UNet(
    input_shape=(64, 64, 4),
    num_blocks=[2, 2, 2, 2],
    hidden_channels=320,
    hidden_expands=[1, 2, 4, 4],
    num_attention_blocks=[1, 1, 1, 0],  # attention_blocks after each res_block in each stack
    num_heads=8,
    mlp_ratio=4,
    conditional_embedding=768,  # > 0 value for using text conditional as generating instruction.
    num_classes=0,  # > 0 value for also using labels as generating instruction.
    activation="swish",
    dropout=0,
    pretrained="v1_5",
    model_name="unet",
    kwargs=None,  # Not using, recieving parameter
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    inputs = layers.Input(backend.align_input_shape_by_image_data_format(input_shape), name="inputs")
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    time_steps = layers.Input([], dtype="int64", name="time_steps")
    condition = layers.Input([None, conditional_embedding], name="condition") if conditional_embedding > 0 else None

    time_embedding = SinusoidalTimeStepEmbedding(hidden_channels=hidden_channels, name="time_embedding")(time_steps)
    time_embedding = layers.Dense(hidden_channels * 4, name="time_embed_1_dense")(time_embedding)
    time_embedding = activation_by_name(time_embedding, activation=activation, name="time_embed_")
    time_embedding = layers.Dense(hidden_channels * 4, name="time_embed_2_dense")(time_embedding)

    if num_classes > 0:
        labels_inputs = layers.Input([], dtype="int64", name="labels_inputs")
        labels_embedding = layers.Embedding(num_classes + 1, hidden_channels, mask_zero=True, name="labels_embedding")(labels_inputs)
        labels_embedding = layers.Dense(hidden_channels * 4, name="labels_embed_1_dense")(labels_embedding)
        labels_embedding = activation_by_name(labels_embedding, activation=activation, name="labels_embed_")
        labels_embedding = layers.Dense(hidden_channels * 4, name="labels_embed_2_dense")(labels_embedding)
    else:
        labels_embedding = None

    nn = conv2d_no_bias(inputs, hidden_channels, kernel_size=3, use_bias=True, padding="SAME", name="latents_")

    """ Down blocks """
    skip_connections = [nn]
    for stack_id, (num_block, hidden_expand, num_attention_block) in enumerate(zip(num_blocks, hidden_expands, num_attention_blocks)):
        stack_name = "stack{}_".format(stack_id + 1)
        out_channels = hidden_expand * hidden_channels
        if stack_id > 0:
            nn = conv2d_no_bias(nn, nn.shape[channel_axis], kernel_size=3, strides=2, use_bias=True, padding="SAME", name=stack_name + "downsample_")
            skip_connections.append(nn)

        for block_id in range(num_block):
            block_name = stack_name + "down_block{}_".format(block_id + 1)
            nn = res_block(nn, time_embedding, labels_embedding, out_channels=out_channels, activation=activation, dropout=dropout, name=block_name)
            if num_attention_block > 0:
                nn = spatial_transformer_block(nn, condition, num_attention_block, mlp_ratio=mlp_ratio, num_heads=num_heads, name=block_name + "attn_")
            skip_connections.append(nn)
    # print(f">>>> {[ii.shape for ii in skip_connections] = }")

    """ Middle blocks """
    nn = res_block(nn, time_embedding, labels_embedding=labels_embedding, activation=activation, dropout=dropout, name="middle_block_1_")
    nn = spatial_transformer_block(nn, condition, num_attention_block=1, name="middle_block_attn_")
    nn = res_block(nn, time_embedding, labels_embedding=labels_embedding, activation=activation, dropout=dropout, name="middle_block_2_")

    """ Up blocks """
    for stack_id, (num_block, hidden_expand, num_attention_block) in enumerate(zip(num_blocks[::-1], hidden_expands[::-1], num_attention_blocks[::-1])):
        stack_name = "stack{}_".format(len(num_blocks) + stack_id + 1)
        out_channels = hidden_expand * hidden_channels
        if stack_id > 0:
            nn = layers.UpSampling2D(size=2, name=stack_name + "upsample_")(nn)
            nn = conv2d_no_bias(nn, nn.shape[channel_axis], kernel_size=3, strides=1, use_bias=True, padding="SAME", name=stack_name + "upsample_")

        for block_id in range(num_block + 1):
            block_name = stack_name + "up_block{}_".format(block_id + 1)
            skip_connection = skip_connections.pop(-1)
            nn = functional.concat([nn, skip_connection], axis=channel_axis)

            nn = res_block(nn, time_embedding, labels_embedding, out_channels=out_channels, activation=activation, dropout=dropout, name=block_name)
            if num_attention_block > 0:
                nn = spatial_transformer_block(nn, condition, num_attention_block, mlp_ratio=mlp_ratio, num_heads=num_heads, name=block_name + "attn_")

    """ Output blocks """
    nn = group_norm(nn, name="output_")
    nn = activation_by_name(nn, activation=activation, name="output_")
    nn = conv2d_no_bias(nn, inputs.shape[channel_axis], kernel_size=3, use_bias=True, padding="SAME", name="output_")
    outputs = layers.Activation("linear", dtype="float32", name="output")(nn)

    model_inputs = [inputs, labels_inputs, time_steps] if num_classes > 0 else [inputs, time_steps]
    model_inputs += [condition] if conditional_embedding > 0 else []
    model = models.Model(model_inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "stable_diffusion", pretrained)

    model.run_prediction = RunPrediction(model=model)
    return model


@register_model
def UNetTest(input_shape=(32, 32, 3), conditional_embedding=0, num_classes=0, activation="swish", pretrained=None, **kwargs):
    hidden_channels = 128
    hidden_expands = [1, 2, 2, 4]
    num_attention_blocks = [0, 0, 1, 1]
    return UNet(**locals(), model_name="unet_test", **kwargs)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    conv2d_no_bias,
    drop_block,
    mlp_block,
    scaled_dot_product_attention,
    PositionalEmbedding,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "hiera_base": {"mae_in1k_ft1k": {224: "7998153d4cf4a2571e1268c8be1d9324"}},
    "hiera_base_plus": {"mae_in1k_ft1k": {224: "5492209f8d7f8632e5db218e2b32496d"}},
    "hiera_huge": {"mae_in1k_ft1k": {224: "2226812b40266ab28e4394849ba7690f"}},
    "hiera_large": {"mae_in1k_ft1k": {224: "68a3f6329f54f5f3bf7737989e3b67df"}},
    "hiera_small": {"mae_in1k_ft1k": {224: "aaba43ec87328a487a06f24d67a4e56d"}},
    "hiera_tiny": {"mae_in1k_ft1k": {224: "d7b913744fc2371e489f536dc236ae03"}},
}


def mhsa_with_window_extracted_and_strides(
    inputs, num_heads=4, key_dim=0, out_shape=None, window_size_prod=-1, strides_prod=1, qkv_bias=True, out_bias=True, attn_dropout=0, name=None
):
    _, blocks, cc = inputs.shape
    out_shape = cc if out_shape is None else out_shape
    key_dim = key_dim if key_dim > 0 else out_shape // num_heads  # Note: different from others using input_channels
    qkv_out = num_heads * key_dim
    qk_scale = 1.0 / (float(key_dim) ** 0.5)
    window_size_prod = window_size_prod if window_size_prod > 0 else blocks
    window_blocks = blocks // window_size_prod
    # print(f"{blocks = }, {window_blocks = }, {window_size_prod = }, {num_heads = }")

    qkv = layers.Dense(qkv_out * 3, use_bias=qkv_bias, name=name and name + "qkv")(inputs)
    query, key, value = functional.split(qkv, 3, axis=-1)

    if strides_prod > 1:
        query = functional.reshape(query, [-1, strides_prod, window_size_prod // strides_prod, window_blocks * num_heads, key_dim])
        query = functional.reduce_max(query, axis=1)
        query = functional.transpose(query, [0, 2, 1, 3])
    else:
        query = functional.transpose(functional.reshape(query, [-1, window_size_prod, window_blocks * num_heads, key_dim]), [0, 2, 1, 3])
    key = functional.transpose(functional.reshape(key, [-1, window_size_prod, window_blocks * num_heads, key_dim]), [0, 2, 3, 1])
    value = functional.transpose(functional.reshape(value, [-1, window_size_prod, window_blocks * num_heads, key_dim]), [0, 2, 1, 3])

    output_shape = [-1, blocks // strides_prod, out_shape]
    return scaled_dot_product_attention(query, key, value, output_shape, out_bias=out_bias, dropout=attn_dropout, name=name)


def attention_mlp_block(inputs, out_channels=-1, num_heads=4, window_size_prod=-1, strides_prod=1, mlp_ratio=4, drop_rate=0, activation="gelu", name=""):
    # print(f">>>> {inputs.shape = }, {drop_rate = }")
    input_channels = inputs.shape[-1]
    out_channels = out_channels if out_channels > 0 else input_channels

    pre = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "attn_ln")(inputs)  # "channels_first" also using axis=-1
    if strides_prod > 1 or out_channels != input_channels:
        short = pre if out_channels == input_channels else layers.Dense(out_channels, name=name + "short_dense")(pre)
        short = functional.reduce_max(functional.reshape(short, [-1, strides_prod, short.shape[1] // strides_prod, short.shape[-1]]), axis=1)
    else:
        short = inputs

    """ Attention """
    attn = mhsa_with_window_extracted_and_strides(
        pre, num_heads=num_heads, out_shape=out_channels, window_size_prod=window_size_prod, strides_prod=strides_prod, name=name + "attn_"
    )
    attn = drop_block(attn, drop_rate)
    attn_out = layers.Add(name=name + "attn_out")([short, attn])

    """ MLP """
    nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "mlp_ln")(attn_out)  # "channels_first" also using axis=-1
    nn = mlp_block(nn, hidden_dim=int(out_channels * mlp_ratio), activation=activation, name=name + "mlp_")
    nn = drop_block(nn, drop_rate)
    nn = layers.Add(name=name + "mlp_output")([attn_out, nn])
    return nn


def unroll(inputs, strides=[2, 2, 2]):
    """
    inputs: [batch, height, width, channels], strides: [2, 2, 2]
    -> [batch, height // 8, 2_h3, 2_h2, 2_h1, width // 8, 2_w3, 2_w2, 2_w1, channels]
    -> [batch, 2_h1, 2_w1, 2_h2, 2_w2, 2_h3, 2_w3, height // 8, width // 8, channels]  # [0, 4, 8, 3, 7, 2, 6, 1, 5, 9]
    -> [batch, height * width, channels]
    """
    height, width, channels = inputs.shape[1:]
    nn = inputs
    for ii in strides:
        nn = functional.reshape(nn, [-1, nn.shape[-3] // ii, ii, nn.shape[-2] // ii, ii, nn.shape[-1]])
        nn = functional.transpose(nn, [0, 2, 4, 1, 3, 5])
    return functional.reshape(nn, [-1, height * width, channels])

    # height_strided = height // np.prod(strides)
    # width_strided = width // np.prod(strides)
    # inner_shape = [-1, height_strided, *strides, width_strided, *strides, channels]
    # nn = functional.reshape(inputs, inner_shape)
    #
    # strides_len = len(strides) + 1
    # perm = [0] + np.ravel([[ii, ii + strides_len] for ii in range(strides_len, 0, -1)]).tolist() + [2 * strides_len + 1]  # [0, 4, 8, 3, 7, 2, 6, 1, 5, 9]
    # nn = functional.transpose(nn, perm)
    #
    # return functional.reshape(nn, [-1, height * width, channels])


def reroll(inputs, strides, height=-1):
    pass
    # for ii in range(strides):
    #     inputs = functional.reshape(inputs, ii, ii, inputs.shape[1:])


def Hiera(
    num_blocks=[1, 2, 7, 2],
    embed_dim=96,
    num_heads=[1, 2, 4, 8],
    use_window_attentions=[True, True, [True, False], False],  # [True, False] means first one is True, others are False
    mlp_ratio=4,
    strides=[1, 2, 2, 2],
    # window_ratios=[8, 4, 1, 1],
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    dropout=0,
    extract_features=False,
    classifier_activation="softmax",
    pretrained=None,
    model_name="hiera",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    window_size_prod = int(np.prod(strides) ** 2)  # Total downsample rates after stem

    """ forward_embeddings """
    nn = conv2d_no_bias(inputs, embed_dim, 7, strides=4, padding="same", use_bias=True, name="stem_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
    nn = PositionalEmbedding(input_height=nn.shape[1], name="positional_embedding")(nn)
    height, width = nn.shape[1:-1]
    nn = unroll(nn, strides=strides[1:])
    # window_ratios = (window_ratios[0] * window_ratios[1]) if isinstance(window_ratios, (list, tuple)) else window_ratios

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    featrues = []
    for stack_id, (num_block, stride, use_window_attention) in enumerate(zip(num_blocks, strides, use_window_attentions)):
        stack_name = "stack{}_".format(stack_id + 1)
        cur_out_channels = embed_dim * (2**stack_id)
        cur_num_heads = num_heads[stack_id] if isinstance(num_heads, (list, tuple)) else num_heads
        stack_use_window_attention = use_window_attention if isinstance(use_window_attention, (list, tuple)) else [use_window_attention]
        stack_use_window_attention = stack_use_window_attention + stack_use_window_attention[-1:] * (num_block - len(stack_use_window_attention))
        window_size_prod //= stride**2
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            strides_prod = (stride**2) if block_id == 0 else 1
            cur_window_size_prod = int(window_size_prod * strides_prod) if stack_use_window_attention[block_id] else -1
            nn = attention_mlp_block(
                nn, cur_out_channels, cur_num_heads, cur_window_size_prod, strides_prod, mlp_ratio, block_drop_rate, activation=activation, name=block_name
            )
            global_block_id += 1
        height, width = height // stride, width // stride  # [TODO] reroll

    nn = functional.reshape(nn, [-1, height, width, nn.shape[-1]])
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name="post_ln")(nn)
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    if extract_features:
        model.extract_features = lambda: features
    reload_model_weights(model, PRETRAINED_DICT, "hiera", pretrained, PositionalEmbedding)
    return model


@register_model
def HieraTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="mae_in1k_ft1k", **kwargs):
    return Hiera(**locals(), model_name="hiera_tiny", **kwargs)


@register_model
def HieraSmall(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="mae_in1k_ft1k", **kwargs):
    num_blocks = [1, 2, 11, 2]
    return Hiera(**locals(), model_name="hiera_small", **kwargs)


@register_model
def HieraBase(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="mae_in1k_ft1k", **kwargs):
    num_blocks = [2, 3, 16, 3]
    return Hiera(**locals(), model_name="hiera_base", **kwargs)


@register_model
def HieraBasePlus(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="mae_in1k_ft1k", **kwargs):
    num_blocks = [2, 3, 16, 3]
    embed_dim = 112
    num_heads = [2, 4, 8, 16]
    return Hiera(**locals(), model_name="hiera_base_plus", **kwargs)


@register_model
def HieraLarge(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="mae_in1k_ft1k", **kwargs):
    num_blocks = [2, 6, 36, 4]
    embed_dim = 144
    num_heads = [2, 4, 8, 16]
    return Hiera(**locals(), model_name="hiera_large", **kwargs)


@register_model
def HieraHuge(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="mae_in1k_ft1k", **kwargs):
    num_blocks = [2, 6, 36, 4]
    embed_dim = 256
    num_heads = [4, 8, 16, 32]
    return Hiera(**locals(), model_name="hiera_huge", **kwargs)



from keras_cv_attention_models.beit.beit import Beit, keras_model_load_weights_from_pytorch_model
from keras_cv_attention_models.models import register_model


def DINOv2(layer_scale=1.0, use_abs_pos_emb=True, use_cat_head=True, attn_qkv_bias=True, model_name="dinov2", **kwargs):
    kwargs.pop("kwargs", None)
    patch_size = kwargs.pop("patch_size", 14)
    force_reload_mismatch = patch_size != 14  # If patch_size not 14, force reload pos_emb and stem_conv weights
    return Beit(**locals(), **kwargs)


@register_model
def DINOv2_ViT_Small14(input_shape=(518, 518, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 384
    depth = 12
    num_heads = 6
    return DINOv2(**locals(), model_name="dinov2_vit_small14", **kwargs)


@register_model
def DINOv2_ViT_Base14(input_shape=(518, 518, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 768
    depth = 12
    num_heads = 12
    return DINOv2(**locals(), model_name="dinov2_vit_base14", **kwargs)


@register_model
def DINOv2_ViT_Large14(input_shape=(518, 518, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 1024
    depth = 24
    num_heads = 16
    return DINOv2(**locals(), model_name="dinov2_vit_large14", **kwargs)


@register_model
def DINOv2_ViT_Giant14(input_shape=(518, 518, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 1536
    depth = 40
    num_heads = 24
    use_gated_mlp = True
    mlp_ratio = 4096 / 1536
    return DINOv2(**locals(), model_name="dinov2_vit_giant14", **kwargs)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format, initializers
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    ChannelAffine,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    mlp_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "edgenext_base": {
        "imagenet": {256: "9492ab3b407aabe84cead0bded7f7dd7"},
        "usi": {256: "9d93a7524eda4701b2364540a2a44c8a"},
        "imagenet21k-ft1k": {256: "d82a6f8cfd1e85cb16b0d6ced189eb1e"},
    },
    "edgenext_small": {"imagenet": {256: "0234641a703283de1cb0d935bb0325e4"}, "usi": {256: "c237761b5bd5c32041d6b758186a0716"}},
    "edgenext_x_small": {"imagenet": {256: "472df7659422c7feffbec8012a0f6fa4"}},
    "edgenext_xx_small": {"imagenet": {256: "4190ba28c7caa2fe73215448f8abebd6"}},
}
LAYER_NORM_EPSILON = 1e-6


@backend.register_keras_serializable(package="kecam/edgenext")
class PositionalEncodingFourier(layers.Layer):
    def __init__(self, filters=32, temperature=1e4, **kwargs):
        super().__init__(**kwargs)
        self.filters, self.temperature = filters, float(temperature)
        self.epsilon = 1e-6
        self.scale = 2 * np.math.acos(-1.0)  # 2 * pi

    def build(self, input_shape):
        _, height, width, channels = input_shape  # ex: height, width, filters = 12, 27, 32
        hh, ww = np.arange(height, dtype="float32"), np.arange(width, dtype="float32")
        hh = (hh + 1) / (float(height) + self.epsilon) * self.scale
        ww = (ww + 1) / (float(width) + self.epsilon) * self.scale

        dim_t = self.temperature ** (2 * (np.arange(self.filters, dtype="float32") // 2) / self.filters)  # (filters,)
        pos_hh, pos_ww = np.expand_dims(hh, -1) / dim_t, np.expand_dims(ww, -1) / dim_t  # pos_hh [12, 32], pos_ww [27, 32]
        pos_hh = np.stack([np.sin(pos_hh[:, 0::2]), np.cos(pos_hh[:, 1::2])], axis=-1)  # pos_hh [12, 16, 2]
        pos_ww = np.stack([np.sin(pos_ww[:, 0::2]), np.cos(pos_ww[:, 1::2])], axis=-1)  # pos_ww [27, 16, 2]
        pos_hh = np.repeat(np.reshape(pos_hh, [height, 1, -1]), width, axis=1)  # [12, 27, 32]
        pos_ww = np.repeat(np.reshape(pos_ww, [1, width, -1]), height, axis=0)  # [12, 27, 32]
        positional_embedding = np.concatenate([pos_hh, pos_ww], axis=-1)  # [12, 27, 64]

        if hasattr(self, "register_buffer"):  # PyTorch
            self.register_buffer("positional_embedding", functional.convert_to_tensor(positional_embedding, dtype=self.compute_dtype), persistent=False)
        else:
            self.positional_embedding = functional.convert_to_tensor(positional_embedding, dtype=self.compute_dtype)

        self.token_projection_ww = self.add_weight(name="ww", initializer="glorot_uniform", shape=(self.filters * 2, channels), trainable=True)
        self.token_projection_bb = self.add_weight(name="bb", initializer="zeros", shape=(channels,), trainable=True)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        pos_emb = self.positional_embedding @ self.token_projection_ww + self.token_projection_bb
        # tf.print(pos_emb.shape, attention_scores.shape)
        return inputs + pos_emb

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"filters": self.filters, "temperature": self.temperature})
        return base_config


def norm_inverted_bottleneck(inputs, mlp_ratio=4, layer_scale=1e-6, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1]  # channels_last only, it should be permuted before entering this
    nn = layer_norm(inputs, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name)
    nn = mlp_block(nn, input_channel * mlp_ratio, activation=activation, name=name)
    nn = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=-1, name=name + "gamma")(nn) if layer_scale >= 0 else nn
    nn = drop_block(nn, drop_rate=drop_rate, name=name)
    return nn


def cross_covariance_attention(inputs, num_heads=4, key_dim=0, qkv_bias=True, out_bias=True, attn_dropout=0, out_dropout=0, name=None):
    input_channel = inputs.shape[-1]  # channels_last only, it should be permuted before entering this
    input_blocks = inputs.shape[1:-1]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    qk_out = key_dim * num_heads

    qkv = functional.reshape(inputs, [-1, int(np.prod(input_blocks)), inputs.shape[-1]]) if len(inputs.shape) > 3 else inputs
    qkv = layers.Dense(qk_out * 3, use_bias=True, name=name and name + "qkv")(qkv)
    query, key, value = functional.split(qkv, 3, axis=-1)
    query = functional.transpose(functional.reshape(query, [-1, query.shape[1], num_heads, key_dim]), [0, 2, 3, 1])  #  [batch, num_heads, key_dim, hh * ww]
    key = functional.transpose(functional.reshape(key, [-1, key.shape[1], num_heads, key_dim]), [0, 2, 1, 3])  # [batch, num_heads, hh * ww, key_dim]
    value = functional.transpose(functional.reshape(value, [-1, value.shape[1], num_heads, key_dim]), [0, 2, 3, 1])  # [batch, num_heads, key_dim, hh * ww]

    norm_query, norm_key = functional.l2_normalize(query, axis=-1, epsilon=1e-6), functional.l2_normalize(key, axis=-2, epsilon=1e-6)
    attn = functional.matmul(norm_query, norm_key)  # [batch, num_heads, key_dim, key_dim]
    attn = ChannelAffine(axis=1, use_bias=False, name=name and name + "temperature/no_weight_decay")(attn)  # axis=1 means on head dimension
    attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attn)

    if attn_dropout > 0:
        attention_scores = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(attention_scores)
    # [batch, num_heads, key_dim, key_dim] * [batch, num_heads, key_dim, hh * ww] -> [batch, num_heads, key_dim, hh * ww]
    attention_output = functional.matmul(attention_scores, value)
    attention_output = functional.transpose(attention_output, [0, 3, 1, 2])  # [batch, hh * ww, num_heads, key_dim]
    attention_output = functional.reshape(attention_output, [-1, *input_blocks, num_heads * key_dim])  # [batch, hh, ww, num_heads * key_dim]
    # print(f">>>> {attention_output.shape = }, {attention_scores.shape = }")

    # [batch, hh, ww, num_heads * key_dim] * [num_heads * key_dim, out] --> [batch, hh, ww, out]
    attention_output = layers.Dense(qk_out, use_bias=out_bias, name=name and name + "output")(attention_output)
    attention_output = layers.Dropout(out_dropout, name=name and name + "out_drop")(attention_output) if out_dropout > 0 else attention_output
    return attention_output


def split_depthwise_transpose_attention(
    inputs, split=1, num_heads=4, mlp_ratio=4, use_pos_emb=False, layer_scale=1e-6, drop_rate=0, activation="gelu", name=""
):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    sub_channels = int(np.math.ceil(input_channel / split))

    if image_data_format() == "channels_last":
        spx, remainder = inputs[:, :, :, : (split - 1) * sub_channels], inputs[:, :, :, (split - 1) * sub_channels :]
    else:
        spx, remainder = inputs[:, : (split - 1) * sub_channels], inputs[:, (split - 1) * sub_channels :]
    spx = functional.split(spx, split - 1, axis=channel_axis)
    gathered_result = []
    for id, ii in enumerate(spx):
        sp = ii if id == 0 else (sp + ii)
        sp = depthwise_conv2d_no_bias(sp, kernel_size=3, padding="same", use_bias=True, name=name + "spx_{}_".format(id + 1))
        gathered_result.append(sp)
    gathered_result.append(remainder)
    attn = functional.concat(gathered_result, axis=channel_axis)
    # print(f"{inputs.shape = }, {attn.shape = }")

    # XCA
    attn = attn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(attn)  # channels_first -> channels_last
    attn = PositionalEncodingFourier(name=name + "pos")(attn) if use_pos_emb else attn

    attn_height, attn_width = attn.shape[1:-1]
    attn = functional.reshape(attn, [-1, attn_height * attn_width, attn.shape[-1]])  # Using 3D for attention inputs

    nn = layer_norm(attn, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "xca_")
    nn = cross_covariance_attention(nn, num_heads, name=name + "xca_")
    nn = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=-1, name=name + "xca_gamma")(nn) if layer_scale >= 0 else nn
    nn = drop_block(nn, drop_rate=drop_rate, name=name + "xca_")
    nn = layers.Add(name=name + "xca")([attn, nn])

    # Inverted Bottleneck
    nn = norm_inverted_bottleneck(nn, mlp_ratio, layer_scale, drop_rate, activation=activation, name=name + "ir_")
    nn = functional.reshape(nn, [-1, attn_height, attn_width, nn.shape[-1]])  # Revert 3D to 4D
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first
    return layers.Add(name=name + "output")([inputs, nn])


def conv_encoder(inputs, mlp_ratio=4, kernel_size=7, layer_scale=1e-6, drop_rate=0, activation="gelu", name=""):
    nn = depthwise_conv2d_no_bias(inputs, kernel_size, use_bias=True, padding="same", name=name)
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
    nn = norm_inverted_bottleneck(nn, mlp_ratio, layer_scale, drop_rate, activation=activation, name=name)
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first
    # print(f"{nn.shape = }, {inputs.shape = }")
    return layers.Add(name=name + "output")([inputs, nn])


def EdgeNeXt(
    num_blocks=[2, 2, 6, 2],
    out_channels=[24, 48, 88, 168],
    num_heads=4,
    num_stda_layers=[0, 1, 1, 1],
    stda_split=[2, 2, 3, 4],
    stda_use_pos_emb=[False, True, False, False],
    conv_kernel_size=[3, 5, 7, 9],
    stem_width=-1,
    mlp_ratio=4,
    stem_patch_size=4,
    layer_scale=1e-6,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="edgenext",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = conv2d_no_bias(inputs, stem_width, kernel_size=stem_patch_size, strides=stem_patch_size, use_bias=True, padding="valid", name="stem_")
    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="stem_")

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, num_stda_layer) in enumerate(zip(num_blocks, out_channels, num_stda_layers)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            ds_name = stack_name + "downsample_"
            nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name=ds_name)
            # Set use_torch_padding=False, as kernel_size == 2, otherwise shape will be enlarged by 1
            nn = conv2d_no_bias(nn, out_channel, kernel_size=2, strides=2, use_bias=True, padding="valid", name=ds_name)
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            if block_id > num_block - num_stda_layer - 1:
                split = stda_split[stack_id]
                use_pos_emb = stda_use_pos_emb[stack_id]
                num_head = num_heads[stack_id] if isinstance(num_heads, (list, tuple)) else num_heads
                nn = split_depthwise_transpose_attention(
                    nn, split, num_head, mlp_ratio, use_pos_emb, layer_scale, block_drop_rate, activation, name=block_name + "stda_"
                )
            else:
                kernel_size = conv_kernel_size[stack_id]
                nn = conv_encoder(nn, mlp_ratio, kernel_size, layer_scale, block_drop_rate, activation=activation, name=block_name + "conv_")
            global_block_id += 1

    """ output """
    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="pre_output_")
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "edgenext", pretrained)
    return model


@register_model
def EdgeNeXt_XX_Small(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return EdgeNeXt(**locals(), model_name="edgenext_xx_small", **kwargs)


@register_model
def EdgeNeXt_X_Small(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 3]
    out_channels = [32, 64, 100, 192]
    return EdgeNeXt(**locals(), model_name="edgenext_x_small", **kwargs)


@register_model
def EdgeNeXt_Small(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 3]
    out_channels = [48, 96, 160, 304]
    num_heads = 8
    return EdgeNeXt(**locals(), model_name="edgenext_small", **kwargs)


@register_model
def EdgeNeXt_Base(input_shape=(256, 256, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 3]
    out_channels = [80, 160, 288, 584]
    num_heads = 8
    return EdgeNeXt(**locals(), model_name="edgenext_base", **kwargs)



import sys
import numpy as np


class Callback:
    def __init__(self):
        self.validation_data = None
        self.model = None

    def set_params(self, params):
        self.params = params

    def set_model(self, model):
        self.model = model

    def on_train_batch_begin(self, batch, logs=None):
        pass

    def on_train_batch_end(self, batch, logs=None):
        pass

    def on_epoch_begin(self, cur_epoch, logs=None):
        pass

    def on_epoch_end(self, cur_epoch, logs=None):
        pass

    def on_test_batch_begin(self, batch, logs=None):
        pass

    def on_test_batch_end(self, batch, logs=None):
        pass

    def on_test_begin(self, logs=None):
        pass

    def on_test_end(self, logs=None):
        pass


class TerminateOnNaN(Callback):
    def on_train_batch_end(self, batch, logs=None):
        logs = logs or {}
        loss = logs.get("loss")
        if loss is not None:
            if not np.isfinite(loss):
                print("\nError: Invalid loss, terminating training")
                # self.model.stop_training = True
                sys.exit()


class TensorBoard(Callback):
    def __init__(self, log_dir="logs", histogram_freq=1, **kwargs):
        super().__init__()
        self.log_dir, self.histogram_freq = log_dir, histogram_freq
        try:
            from torch.utils.tensorboard import SummaryWriter

            self.tensorboard_writer = SummaryWriter(self.log_dir)
            print(">>>> Tensorboard writer created, summary will be write to '{}', view by 'tensorboard --logdir {}'".format(log_dir, log_dir))
        except:
            self.tensorboard_writer = None
            print("[Error] tensorboard not installed, try `pip install tensorboard`")

    def on_epoch_end(self, cur_epoch, logs=None):
        if self.tensorboard_writer is None:
            return
        logs = logs or {}
        for kk, vv in logs.items():
            self.tensorboard_writer.add_scalar(kk, vv, cur_epoch)
